import autofit as af
import autolens as al

import os

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))
output_path = workspace_path + "../../outputs/PyAutoLens/"
af.conf.instance = af.conf.Config(
    config_path=workspace_path + "config", output_path=output_path
)

data_type = "noise_10"

data_level = "level_0"

data_name = "large_hi_sn_system_1"
data_name = "small_hi_sn_system_1"

# data_level = 'level_1'
# data_name = 'large_hi_sn_system_1'
# data_name = 'large_hi_sn_system_2'
# data_name = 'large_md_sn_system_1'
# data_name = 'large_md_sn_system_2'
# data_name = 'large_lo_sn_system_1'
# data_name = 'large_lo_sn_system_2'
# data_name = 'small_hi_sn_system_1'
# data_name = 'small_hi_sn_system_2'
# data_name = 'small_md_sn_system_1'
# data_name = 'small_md_sn_system_2'
# data_name = 'small_lo_sn_system_1'
# data_name = 'small_lo_sn_system_2'

pixel_scale = 0.00976562

data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path,
    folder_names=["data", "subhalo_challenge", data_type, data_level, data_name],
)

ccd_data = al.load_ccd_data_from_fits(
    image_path=data_path + "image.fits",
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    pixel_scale=pixel_scale,
)

mask = al.load_mask_from_fits(
    mask_path=data_path + "mask_irregular.fits", pixel_scale=pixel_scale
)

from workspace_jam.pipelines.advanced.no_lens_light.initialize import (
    lens_sie_shear_source_sersic,
)
from workspace_jam.pipelines.advanced.no_lens_light.power_law.from_initialize import (
    lens_pl_shear_source_sersic,
)
from workspace_jam.pipelines.advanced.no_lens_light.subhalo.from_power_law import (
    lens_pl_shear_subhalo_source_sersic,
)

pipeline_initialize = lens_sie_shear_source_sersic.make_pipeline(
    phase_folders=[data_type, data_level, data_name]
)

pipeline_power_law = lens_pl_shear_source_sersic.make_pipeline(
    phase_folders=[data_type, data_level, data_name]
)

pipeline_subhalo = lens_pl_shear_subhalo_source_sersic.make_pipeline(
    phase_folders=[data_type, data_level, data_name]
)

pipeline = pipeline_initialize + pipeline_power_law + pipeline_subhalo

pipeline.run(data=ccd_data, mask=mask)



import autofit as af
from autolens.pipeline.phase import phase_imaging
from .hyper_galaxy_phase import HyperGalaxyPhase
from .hyper_phase import HyperPhase
from .inversion_phase import InversionBackgroundBothPhase
from .inversion_phase import InversionBackgroundNoisePhase
from .inversion_phase import InversionBackgroundSkyPhase
from .inversion_phase import InversionPhase
from .inversion_phase import VariableFixingHyperPhase


class CombinedHyperPhase(HyperPhase):
    def __init__(
        self, phase: phase_imaging.PhaseImaging, hyper_phase_classes: (type,) = tuple()
    ):
        """
        A hyper_combined hyper_galaxies phase that can run zero or more other hyper_galaxies phases after the initial phase is
        run.

        Parameters
        ----------
        phase : phase_imaging.PhaseImaging
            The phase wrapped by this hyper_galaxies phase
        hyper_phase_classes
            The classes of hyper_galaxies phases to be run following the initial phase
        """
        super().__init__(phase, "hyper_combined")
        self.hyper_phases = list(map(lambda cls: cls(phase), hyper_phase_classes))

    @property
    def phase_names(self) -> [str]:
        """
        The names of phases included in this hyper_combined phase
        """
        return [phase.hyper_name for phase in self.hyper_phases]

    def run(
        self,
        data,
        results: af.ResultsCollection = None,
        mask=None,
        positions=None,
        **kwargs
    ) -> af.Result:
        """
        Run the phase followed by the hyper_galaxies phases. Each result of a hyper_galaxies phase is attached to the
        overall result object by the hyper_name of that phase.

        Finally, a phase in run with all of the variable results from all the individual hyper_galaxies phases.

        Parameters
        ----------
        positions
        mask
        data
            the data
        results
            Results from previous phases
        kwargs

        Returns
        -------
        result
            The result of the phase, with hyper_galaxies results attached by associated hyper_galaxies names
        """

        results = results.copy() if results is not None else af.ResultsCollection()
        result = self.phase.run(
            data, results=results, mask=mask, positions=positions, **kwargs
        )
        results.add(self.phase.phase_name, result)

        for phase in self.hyper_phases:
            hyper_result = phase.run_hyper(data=data, results=results, **kwargs)
            setattr(result, phase.hyper_name, hyper_result)

        setattr(result, self.hyper_name, self.run_hyper(data=data, results=results))
        return result

    def combine_variables(self, result) -> af.ModelMapper:
        """
        Combine the variable objects from all previous results in this hyper_combined hyper_galaxies phase.

        Iterates through the hyper_galaxies names of the included hyper_galaxies phases, extracting a result
        for each name and adding the variable of that result to a new variable.

        Parameters
        ----------
        result
            The last result (with attribute results associated with phases in this phase)

        Returns
        -------
        combined_variable
            A variable object including all variables from results in this phase.
        """
        variable = af.ModelMapper()
        for name in self.phase_names:
            variable += getattr(result, name).variable
        return variable

    def run_hyper(self, data, results, **kwargs) -> af.Result:
        variable = self.combine_variables(results.last)

        phase = self.make_hyper_phase()
        phase.optimizer.phase_tag = ""
        phase.optimizer.variable = variable

        phase.phase_tag = ""

        return phase.run(
            data,
            results=results,
            mask=results.last.mask_2d,
            positions=results.last.positions,
        )



import autolens as al

import numpy as np

from autolens import exc
from autolens.array.util import mask_util


class MockMask(al.Mask):
    def __new__(cls, array, pixel_scale=1.0, *args, **kwargs):

        obj = np.array(array, dtype="bool").view(cls)
        obj.pixel_scale = pixel_scale
        obj.origin = (0.0, 0.0)

        return obj

    def __init__(self, array, pixel_scale=1.0):
        pass

    def blurring_mask_from_psf_shape(self, psf_shape):
        """Compute a blurring mask, which represents all masked pixels whose light will be blurred into unmasked \
        pixels via PSF convolution (see grid.Grid.blurring_grid_from_mask_and_psf_shape).

        Parameters
        ----------
        psf_shape : (int, int)
           The shape of the psf which defines the blurring region (e.al. the shape of the PSF)
        """

        if psf_shape[0] % 2 == 0 or psf_shape[1] % 2 == 0:
            raise exc.MaskException("psf_size of exterior region must be odd")

        blurring_mask = al.mask_util.blurring_mask_from_mask_and_psf_shape(
            self, psf_shape
        )

        return MockMask(array=blurring_mask, pixel_scale=self.pixel_scale)


class MockMask1D(np.ndarray):
    def __new__(cls, shape, pixel_scale=1.0, *args, **kwargs):

        array = np.full(fill_value=False, shape=shape)

        obj = np.array(array, dtype="bool").view(cls)
        obj.pixel_scale = pixel_scale
        obj.origin = (0.0, 0.0)

        return obj



import logging

import numpy as np

from autolens import exc
from autolens.array.util import grid_util, array_util, mask_util, binning_util
from autolens.array.mapping_util import (
    array_mapping_util,
    grid_mapping_util,
    mask_mapping_util,
)
from autolens.array import scaled_array

logging.basicConfig()
logger = logging.getLogger(__name__)


class Mask(scaled_array.ScaledSquarePixelArray):
    """
    A mask represented by an ndarray where True is masked.
    """

    # noinspection PyUnusedLocal
    def __init__(self, array, pixel_scale, origin=(0.0, 0.0)):
        """ A mask, which is applied to a 2D array of hyper_galaxies to extract a set of unmasked image pixels (i.e. mask entry \
        is *False* or 0) which are then fitted in an analysis.
        
        The mask retains the pixel scale of the array and has a centre and origin.
        
        Parameters
        ----------
        array: ndarray
            An array of bools representing the mask.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        origin : (float, float)
            The (y,x) arc-second origin of the mask's coordinate system.
        centre : (float, float)
            The (y,x) arc-second centre of the mask provided it is a standard geometric shape (e.g. a circle).
        """
        # noinspection PyArgumentList
        super(Mask, self).__init__(array=array, pixel_scale=pixel_scale, origin=origin)

    def __array_finalize__(self, obj):
        if hasattr(obj, "pixel_scale"):
            self.pixel_scale = obj.pixel_scale
        if hasattr(obj, "origin"):
            self.origin = obj.origin
        if hasattr(obj, "origin"):
            self.origin = obj.origin

    def __getitem__(self, coords):
        try:
            return super(Mask, self).__getitem__(coords)
        except IndexError:
            return True

    @classmethod
    def unmasked_for_shape_and_pixel_scale(cls, shape, pixel_scale, invert=False):
        """Setup a mask where all pixels are unmasked.

        Parameters
        ----------
        shape : (int, int)
            The (y,x) shape of the mask in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        """
        mask = np.full(tuple(map(lambda d: int(d), shape)), False)
        if invert:
            mask = np.invert(mask)
        return cls(array=mask, pixel_scale=pixel_scale)

    @classmethod
    def circular(
        cls, shape, pixel_scale, radius_arcsec, centre=(0.0, 0.0), invert=False
    ):
        """Setup a mask where unmasked pixels are within a circle of an input arc second radius and centre.

        Parameters
        ----------
        shape: (int, int)
            The (y,x) shape of the mask in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        radius_arcsec : float
            The radius (in arc seconds) of the circle within which pixels unmasked.
        centre: (float, float)
            The centre of the circle used to mask pixels.
        """
        mask = mask_util.mask_circular_from_shape_pixel_scale_and_radius(
            shape=shape,
            pixel_scale=pixel_scale,
            radius_arcsec=radius_arcsec,
            centre=centre,
        )
        if invert:
            mask = np.invert(mask)
        return cls(array=mask.astype("bool"), pixel_scale=pixel_scale)

    @classmethod
    def circular_annular(
        cls,
        shape,
        pixel_scale,
        inner_radius_arcsec,
        outer_radius_arcsec,
        centre=(0.0, 0.0),
        invert=False,
    ):
        """Setup a mask where unmasked pixels are within an annulus of input inner and outer arc second radii and \
         centre.

        Parameters
        ----------
        shape : (int, int)
            The (y,x) shape of the mask in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        inner_radius_arcsec : float
            The radius (in arc seconds) of the inner circle outside of which pixels are unmasked.
        outer_radius_arcsec : float
            The radius (in arc seconds) of the outer circle within which pixels are unmasked.
        centre: (float, float)
            The centre of the annulus used to mask pixels.
        """

        mask = mask_util.mask_circular_annular_from_shape_pixel_scale_and_radii(
            shape=shape,
            pixel_scale=pixel_scale,
            inner_radius_arcsec=inner_radius_arcsec,
            outer_radius_arcsec=outer_radius_arcsec,
            centre=centre,
        )

        if invert:
            mask = np.invert(mask)

        return cls(array=mask.astype("bool"), pixel_scale=pixel_scale)

    @classmethod
    def circular_anti_annular(
        cls,
        shape,
        pixel_scale,
        inner_radius_arcsec,
        outer_radius_arcsec,
        outer_radius_2_arcsec,
        centre=(0.0, 0.0),
        invert=False,
    ):
        """Setup a mask where unmasked pixels are outside an annulus of input inner and outer arc second radii, but \
        within a second outer radius, and at a given centre.

        This mask there has two distinct unmasked regions (an inner circle and outer annulus), with an inner annulus \
        of masked pixels.

        Parameters
        ----------
        shape : (int, int)
            The (y,x) shape of the mask in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        inner_radius_arcsec : float
            The radius (in arc seconds) of the inner circle inside of which pixels are unmasked.
        outer_radius_arcsec : float
            The radius (in arc seconds) of the outer circle within which pixels are masked and outside of which they \
            are unmasked.
        outer_radius_2_arcsec : float
            The radius (in arc seconds) of the second outer circle within which pixels are unmasked and outside of \
            which they masked.
        centre: (float, float)
            The centre of the anti-annulus used to mask pixels.
        """

        mask = mask_util.mask_circular_anti_annular_from_shape_pixel_scale_and_radii(
            shape=shape,
            pixel_scale=pixel_scale,
            inner_radius_arcsec=inner_radius_arcsec,
            outer_radius_arcsec=outer_radius_arcsec,
            outer_radius_2_arcsec=outer_radius_2_arcsec,
            centre=centre,
        )

        if invert:
            mask = np.invert(mask)

        return cls(array=mask.astype("bool"), pixel_scale=pixel_scale)

    @classmethod
    def elliptical(
        cls,
        shape,
        pixel_scale,
        major_axis_radius_arcsec,
        axis_ratio,
        phi,
        centre=(0.0, 0.0),
        invert=False,
    ):
        """ Setup a mask where unmasked pixels are within an ellipse of an input arc second major-axis and centre.

        Parameters
        ----------
        shape: (int, int)
            The (y,x) shape of the mask in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        major_axis_radius_arcsec : float
            The major-axis (in arc seconds) of the ellipse within which pixels are unmasked.
        axis_ratio : float
            The axis-ratio of the ellipse within which pixels are unmasked.
        phi : float
            The rotation angle of the ellipse within which pixels are unmasked, (counter-clockwise from the positive \
             x-axis).
        centre: (float, float)
            The centre of the ellipse used to mask pixels.
        """

        mask = mask_util.mask_elliptical_from_shape_pixel_scale_and_radius(
            shape, pixel_scale, major_axis_radius_arcsec, axis_ratio, phi, centre
        )

        if invert:
            mask = np.invert(mask)

        return cls(array=mask.astype("bool"), pixel_scale=pixel_scale)

    @classmethod
    def elliptical_annular(
        cls,
        shape,
        pixel_scale,
        inner_major_axis_radius_arcsec,
        inner_axis_ratio,
        inner_phi,
        outer_major_axis_radius_arcsec,
        outer_axis_ratio,
        outer_phi,
        centre=(0.0, 0.0),
        invert=False,
    ):
        """Setup a mask where unmasked pixels are within an elliptical annulus of input inner and outer arc second \
        major-axis and centre.

        Parameters
        ----------
        shape: (int, int)
            The (y,x) shape of the mask in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        inner_major_axis_radius_arcsec : float
            The major-axis (in arc seconds) of the inner ellipse within which pixels are masked.
        inner_axis_ratio : float
            The axis-ratio of the inner ellipse within which pixels are masked.
        inner_phi : float
            The rotation angle of the inner ellipse within which pixels are masked, (counter-clockwise from the \
            positive x-axis).
        outer_major_axis_radius_arcsec : float
            The major-axis (in arc seconds) of the outer ellipse within which pixels are unmasked.
        outer_axis_ratio : float
            The axis-ratio of the outer ellipse within which pixels are unmasked.
        outer_phi : float
            The rotation angle of the outer ellipse within which pixels are unmasked, (counter-clockwise from the \
            positive x-axis).
        centre: (float, float)
            The centre of the elliptical annuli used to mask pixels.
        """

        mask = mask_util.mask_elliptical_annular_from_shape_pixel_scale_and_radius(
            shape=shape,
            pixel_scale=pixel_scale,
            inner_major_axis_radius_arcsec=inner_major_axis_radius_arcsec,
            inner_axis_ratio=inner_axis_ratio,
            inner_phi=inner_phi,
            outer_major_axis_radius_arcsec=outer_major_axis_radius_arcsec,
            outer_axis_ratio=outer_axis_ratio,
            outer_phi=outer_phi,
            centre=centre,
        )

        if invert:
            mask = np.invert(mask)

        return cls(array=mask.astype("bool"), pixel_scale=pixel_scale)

    def binned_up_mask_from_mask(self, bin_up_factor):

        binned_up_mask = binning_util.binned_up_mask_2d_from_mask_2d_and_bin_up_factor(
            mask_2d=self, bin_up_factor=bin_up_factor
        )

        return Mask(
            array=binned_up_mask,
            pixel_scale=self.pixel_scale * bin_up_factor,
            origin=self.origin,
        )

    @property
    @array_util.Memoizer()
    def centre(self):
        return grid_util.grid_centre_from_grid_1d(grid_1d=self.masked_grid_1d)

    @property
    def pixels_in_mask(self):
        return int(np.size(self) - np.sum(self))

    @property
    def mask_1d_index_to_mask_2d_index(self):
        """A 1D array of mappings between every unmasked pixel and its 2D pixel coordinates."""
        return self.sub_mask_1d_index_to_sub_mask_2d_index_from_sub_grid_size(
            sub_grid_size=1
        )

    def sub_mask_1d_index_to_sub_mask_2d_index_from_sub_grid_size(self, sub_grid_size):
        """A 1D array of mappings between every unmasked sub pixel and its 2D sub-pixel coordinates."""
        return mask_mapping_util.sub_mask_1d_index_to_sub_mask_2d_index_from_mask_and_sub_grid_size(
            mask=self, sub_grid_size=sub_grid_size
        ).astype(
            "int"
        )

    def array_1d_from_array_2d(self, array_2d):
        """For a 2D array (e.g. an image, noise_map, etc.) map it to a masked 1D array of valuees using this mask.

        Parameters
        ----------
        array_2d : ndarray | None | float
            The 2D array to be mapped to a masked 1D array.
        """
        if array_2d is None or isinstance(array_2d, float):
            return array_2d
        return array_mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            mask=self, sub_array_2d=array_2d, sub_grid_size=1
        )

    def array_2d_from_array_1d(self, array_1d):
        """ Map a 1D array the same dimension as the grid to its original 2D array.

        Values which were masked in the mapping_util to the 1D array are returned as zeros.

        Parameters
        -----------
        array_1d : ndarray
            The 1D array which is mapped to its masked 2D array.
        """
        return array_mapping_util.sub_array_2d_from_sub_array_1d_mask_and_sub_grid_size(
            sub_array_1d=array_1d, mask=self, sub_grid_size=1
        )

    def scaled_array_2d_from_array_1d(self, array_1d):
        """ Map a 1D array the same dimension as the grid to its original masked 2D array and return it as a hyper \
        array.

        Parameters
        -----------
        array_1d : ndarray
            The 1D array of which is mapped to a 2D hyper array.
        """
        return scaled_array.ScaledSquarePixelArray(
            array=self.array_2d_from_array_1d(array_1d=array_1d),
            pixel_scale=self.pixel_scale,
            origin=self.origin,
        )

    def grid_2d_from_grid_1d(self, grid_1d):
        """Map a 1D grid the same dimension as the grid to its original 2D grid.

        Values which were masked in the mapping_util to the 1D array are returned as zeros.

        Parameters
        -----------
        grid_1d : ndarray
            The 1D grid which is mapped to its masked 2D array.
        """
        return grid_mapping_util.sub_grid_2d_from_sub_grid_1d_mask_and_sub_grid_size(
            sub_grid_1d=grid_1d, mask=self, sub_grid_size=1
        )

    def grid_1d_from_grid_2d(self, grid_2d):
        """ Map a 2D grid to its masked 1D grid..

        Values which are masked in the mapping_util to the 1D grid are returned as zeros.

        Parameters
        -----------
        grid_1d : ndgrid
            The 1D grid which is mapped to its masked 2D grid.
        """
        return grid_mapping_util.sub_grid_1d_from_sub_grid_2d_mask_and_sub_grid_size(
            sub_grid_2d=grid_2d, mask=self, sub_grid_size=1
        )

    def sub_array_2d_from_sub_array_1d_and_sub_grid_size(
        self, sub_array_1d, sub_grid_size
    ):
        """ Map a 1D sub-array the same dimension as the sub-grid (e.g. including sub-pixels) to its original masked
        2D sub array.

        Parameters
        -----------
        sub_array_1d : ndarray
            The 1D sub_array which is mapped to its masked 2D sub-array.
        """
        return array_mapping_util.sub_array_2d_from_sub_array_1d_mask_and_sub_grid_size(
            sub_array_1d=sub_array_1d, mask=self, sub_grid_size=sub_grid_size
        )

    def scaled_array_2d_with_sub_dimensions_from_sub_array_1d_and_sub_grid_size(
        self, sub_array_1d, sub_grid_size
    ):
        """ Map a 1D sub-array the same dimension as the sub-grid to its original masked 2D sub-array and return it as
        a hyper array.

        Parameters
        -----------
        sub_array_1d : ndarray
            The 1D sub-array of which is mapped to a 2D hyper sub-array the dimensions.
        """
        return scaled_array.ScaledSquarePixelArray(
            array=self.sub_array_2d_from_sub_array_1d_and_sub_grid_size(
                sub_array_1d=sub_array_1d, sub_grid_size=sub_grid_size
            ),
            pixel_scale=self.pixel_scale / sub_grid_size,
            origin=self.origin,
        )

    def array_1d_binned_from_sub_array_1d_and_sub_grid_size(
        self, sub_array_1d, sub_grid_size
    ):
        """For an input 1D sub-array, map its values to a 1D array of values by summing each set \of sub-pixel \
        values and dividing by the total number of sub-pixels.

        Parameters
        -----------
        sub_array_1d : ndarray
            A 1D sub-array of values (e.g. image, convergence, potential) which is mapped to
            a 1d array.
        """

        sub_grid_length = int(sub_grid_size ** 2.0)
        sub_grid_fraction = 1.0 / sub_grid_length

        return np.multiply(
            sub_grid_fraction, sub_array_1d.reshape(-1, sub_grid_length).sum(axis=1)
        )

    def scaled_array_2d_binned_from_sub_array_1d_and_sub_grid_size(
        self, sub_array_1d, sub_grid_size
    ):
        """ Map a 1D sub-array the same dimension as the sub-grid to its original masked 2D sub-array and return it as
        a hyper array.

        Parameters
        -----------
        sub_array_1d : ndarray
            The 1D sub-array of which is mapped to a 2D hyper sub-array the dimensions.
        """

        array_1d = self.array_1d_binned_from_sub_array_1d_and_sub_grid_size(
            sub_array_1d=sub_array_1d, sub_grid_size=sub_grid_size
        )

        return scaled_array.ScaledSquarePixelArray(
            array=self.array_2d_from_array_1d(array_1d=array_1d),
            pixel_scale=self.pixel_scale,
            origin=self.origin,
        )

    def sub_array_1d_with_sub_dimensions_from_sub_array_2d_and_sub_grid_size(
        self, sub_array_2d, sub_grid_size
    ):
        """ Map a 2D sub-array to its masked 1D sub-array.

        Values which are masked in the mapping_util to the 1D array are returned as zeros.

        Parameters
        -----------
        su_array_2d : ndarray
            The 2D sub-array which is mapped to its masked 1D sub-array.
        """
        return array_mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            sub_array_2d=sub_array_2d, mask=self, sub_grid_size=sub_grid_size
        )

    def sub_grid_1d_with_sub_dimensions_from_sub_grid_2d_and_sub_grid_size(
        self, sub_grid_2d, sub_grid_size
    ):
        """For an input 1D sub-array, map its values to a 1D array of values by summing each set \of sub-pixel \
        values and dividing by the total number of sub-pixels.

        Parameters
        -----------
        sub_grid_2d : ndarray
            A 1D sub-array of values (e.g. image, convergence, potential) which is mapped to
            a 1d array.
        """
        return grid_mapping_util.sub_grid_1d_from_sub_grid_2d_mask_and_sub_grid_size(
            sub_grid_2d=sub_grid_2d, mask=self, sub_grid_size=sub_grid_size
        )

    def grid_1d_binned_from_sub_grid_1d_and_sub_grid_size(
        self, sub_grid_1d, sub_grid_size
    ):
        """For an input 1D sub-array, map its values to a 1D array of values by summing each set \of sub-pixel \
        values and dividing by the total number of sub-pixels.

        Parameters
        -----------
        sub_grid_1d : ndarray
            A 1D sub-array of values (e.g. image, convergence, potential) which is mapped to
            a 1d array.
        """

        grid_1d_y = self.array_1d_binned_from_sub_array_1d_and_sub_grid_size(
            sub_array_1d=sub_grid_1d[:, 0], sub_grid_size=sub_grid_size
        )

        grid_1d_x = self.array_1d_binned_from_sub_array_1d_and_sub_grid_size(
            sub_array_1d=sub_grid_1d[:, 1], sub_grid_size=sub_grid_size
        )

        return np.stack((grid_1d_y, grid_1d_x), axis=-1)

    def grid_2d_binned_from_sub_grid_1d_and_sub_grid_size(
        self, sub_grid_1d, sub_grid_size
    ):
        """For an input 1D sub-array, map its values to a 1D array of values by summing each set \of sub-pixel \
        values and dividing by the total number of sub-pixels.

        Parameters
        -----------
        sub_grid_1d : ndarray
            A 1D sub-array of values (e.g. image, convergence, potential) which is mapped to
            a 1d array.
        """

        grid_1d = self.grid_1d_binned_from_sub_grid_1d_and_sub_grid_size(
            sub_grid_1d=sub_grid_1d, sub_grid_size=sub_grid_size
        )

        return self.grid_2d_from_grid_1d(grid_1d=grid_1d)

    def sub_grid_2d_with_sub_dimensions_from_sub_grid_1d_and_sub_grid_size(
        self, sub_grid_1d, sub_grid_size
    ):
        """ Map a 1D sub-grid the same dimension as the sub-grid (e.g. including sub-pixels) to its original masked
        2D sub grid.

        Parameters
        -----------
        sub_grid_1d : ndgrid
            The 1D grid which is mapped to its masked 2D sub-grid.
        """
        return grid_mapping_util.sub_grid_2d_from_sub_grid_1d_mask_and_sub_grid_size(
            sub_grid_1d=sub_grid_1d, mask=self, sub_grid_size=sub_grid_size
        )

    @array_util.Memoizer()
    def sub_mask_1d_index_to_mask_1d_index_from_sub_grid_size(self, sub_grid_size):
        """The mapping_util between every sub-pixel and its host pixel.

        For example:

        - sub_to_pixel[8] = 2 -  The ninth sub-pixel is within the 3rd pixel.
        - sub_to_pixel[20] = 4 -  The twenty first sub-pixel is within the 5th pixel.
        """
        return mask_mapping_util.sub_mask_1d_index_to_mask_1d_index_from_mask(
            mask=self, sub_grid_size=sub_grid_size
        ).astype("int")

    @array_util.Memoizer()
    def blurring_mask_from_psf_shape(self, psf_shape):
        """Compute a blurring mask, which represents all masked pixels whose light will be blurred into unmasked \
        pixels via PSF convolution (see grid.Grid.blurring_grid_from_mask_and_psf_shape).

        Parameters
        ----------
        psf_shape : (int, int)
           The shape of the psf which defines the blurring region (e.g. the shape of the PSF)
        """

        if psf_shape[0] % 2 == 0 or psf_shape[1] % 2 == 0:
            raise exc.MaskException("psf_size of exterior region must be odd")

        blurring_mask = mask_util.blurring_mask_from_mask_and_psf_shape(self, psf_shape)

        return Mask(array=blurring_mask, pixel_scale=self.pixel_scale)

    @property
    def edge_1d_indexes(self):
        """The indicies of the mask's edge pixels, where an edge pixel is any unmasked pixel on its edge \
        (next to at least one pixel with a *True* value).
        """
        return mask_util.edge_1d_indexes_from_mask(mask=self).astype("int")

    @property
    def border_1d_indexes(self):
        """The indicies of the mask's border pixels, where a border pixel is any unmasked pixel on an
        exterior edge (e.g. next to at least one pixel with a *True* value but not central pixels like those within \
        an annulus mask).
        """
        return mask_util.border_1d_indexes_from_mask(mask=self).astype("int")

    @property
    def border_grid_1d(self):
        """The indicies of the mask's border pixels, where a border pixel is any unmasked pixel on an
        exterior edge (e.g. next to at least one pixel with a *True* value but not central pixels like those within \
        an annulus mask).
        """
        return self.masked_grid_1d[self.border_1d_indexes]

    def sub_border_1d_indexes_from_sub_grid_size(self, sub_grid_size, filter_size=None):
        """The indicies of the mask's border pixels, where a border pixel is any unmasked pixel on an
        exterior edge (e.g. next to at least one pixel with a *True* value but not central pixels like those within \
        an annulus mask).
        """
        sub_border_1d_indexes = mask_util.sub_border_pixel_1d_indexes_from_mask_and_sub_grid_size(
            mask=self, sub_grid_size=sub_grid_size
        ).astype(
            "int"
        )

        if filter_size is None:
            return sub_border_1d_indexes
        else:
            sub_border_thetas_1d = self.sub_border_thetas_1d_from_sub_grid_size(
                sub_grid_size=sub_grid_size
            )
            sub_border_ordered_indexes = np.argsort(sub_border_thetas_1d)
            return np.sort(
                sub_border_1d_indexes[sub_border_ordered_indexes[1::filter_size]]
            )

    def sub_border_grid_1d_from_sub_grid_size(self, sub_grid_size, filter_size=None):
        """The indicies of the mask's border pixels, where a border pixel is any unmasked pixel on an
        exterior edge (e.g. next to at least one pixel with a *True* value but not central pixels like those within \
        an annulus mask).
        """
        masked_sub_grid_1d = self.masked_sub_grid_1d_from_sub_grid_size(
            sub_grid_size=sub_grid_size
        )
        sub_border_pixel_1d_indexes = self.sub_border_1d_indexes_from_sub_grid_size(
            sub_grid_size=sub_grid_size, filter_size=filter_size
        )
        return masked_sub_grid_1d[sub_border_pixel_1d_indexes]

    def sub_border_thetas_1d_from_sub_grid_size(self, sub_grid_size, filter_size=None):
        sub_border_grid_1d = self.sub_border_grid_1d_from_sub_grid_size(
            sub_grid_size=sub_grid_size, filter_size=filter_size
        )
        return np.arctan2(
            (sub_border_grid_1d[:, 0] - self.centre[0]),
            sub_border_grid_1d[:, 1] - self.centre[1],
        )

    @property
    def masked_grid_1d(self):
        return grid_util.grid_1d_from_mask_pixel_scales_sub_grid_size_and_origin(
            mask=self,
            pixel_scales=self.pixel_scales,
            sub_grid_size=1,
            origin=self.origin,
        )

    def masked_sub_grid_1d_from_sub_grid_size(self, sub_grid_size):
        return grid_util.grid_1d_from_mask_pixel_scales_sub_grid_size_and_origin(
            mask=self,
            pixel_scales=self.pixel_scales,
            sub_grid_size=sub_grid_size,
            origin=self.origin,
        )

    @property
    def zoom_centre(self):
        extraction_grid_1d = self.grid_arcsec_to_grid_pixels(
            grid_arcsec=self.masked_grid_1d
        )
        y_pixels_max = np.max(extraction_grid_1d[:, 0])
        y_pixels_min = np.min(extraction_grid_1d[:, 0])
        x_pixels_max = np.max(extraction_grid_1d[:, 1])
        x_pixels_min = np.min(extraction_grid_1d[:, 1])
        return (
            ((y_pixels_max + y_pixels_min - 1.0) / 2.0),
            ((x_pixels_max + x_pixels_min - 1.0) / 2.0),
        )

    @property
    def zoom_offset_pixels(self):
        return (
            self.zoom_centre[0] - self.central_pixel_coordinates[0],
            self.zoom_centre[1] - self.central_pixel_coordinates[1],
        )

    @property
    def zoom_offset_arcsec(self):
        return (
            -self.pixel_scale * self.zoom_offset_pixels[0],
            self.pixel_scale * self.zoom_offset_pixels[1],
        )

    @property
    def zoom_region(self):
        """The zoomed rectangular region corresponding to the square encompassing all unmasked values. This zoomed
        extraction region is a squuare, even if the mask is rectangular.

        This is used to zoom in on the region of an image that is used in an analysis for visualization."""

        # Have to convert mask to bool for invert function to work.
        where = np.array(np.where(np.invert(self.astype("bool"))))
        y0, x0 = np.amin(where, axis=1)
        y1, x1 = np.amax(where, axis=1)

        ylength = y1 - y0
        xlength = x1 - x0

        if ylength > xlength:
            length_difference = ylength - xlength
            x1 += int(length_difference / 2.0)
            x0 -= int(length_difference / 2.0)
        elif xlength > ylength:
            length_difference = xlength - ylength
            y1 += int(length_difference / 2.0)
            y0 -= int(length_difference / 2.0)

        return [y0, y1 + 1, x0, x1 + 1]


def load_mask_from_fits(mask_path, pixel_scale, mask_hdu=0):
    return Mask.from_fits_with_pixel_scale(
        file_path=mask_path, hdu=mask_hdu, pixel_scale=pixel_scale
    )


def output_mask_to_fits(mask, mask_path, overwrite=False):
    array_util.numpy_array_2d_to_fits(
        array_2d=mask, file_path=mask_path, overwrite=overwrite
    )



import sys
import numpy as np

sys.path.append("../")

from workspace_jam.plotting.density_rj_lens_plots import density_plot_tools

slacs = density_plot_tools.SLACS()

image_dir = (
    "/gpfs/data/pdtw24/PL_Data/RJLens3/"
)  # Dir of Object to make evidence tables from
image_name = "Abell 1201"
file_name = "density_decomposed_aligned"
title = "SIE + Decomposed (Aligned) Density Profiles of "

number_bins = 50

## SIE ##

values = []
values.append(0.05247720)  # Lens x
values.append(0.03497244)  # Lens y
values.append(1.91716393)  # Lens einR
values.append(0.68532604)  # Lens axis ratio
values.append(66.83060729)  # Lens phi

sie_mass_ein_r = slacs.masses_of_all_samples_sie(
    values=values, radius_kpc=slacs.einstein_radius * slacs.kpc_per_arcsec
)
sie_mass_475 = slacs.masses_of_all_samples_sie(values=values, radius_kpc=4.75)
slacs.density_vs_radii_sie(radius_kpc=15.0, values=values, number_bins=number_bins)

### Decomposed Aligned model ###

center_skip = 0
ltm_skip = 0

values = []
values.append(0.02191929)  # fg 1 x
values.append(0.02982811)  # fg 1 y
values.append(0.19665282)  # fg 1 intensity
values.append(0.55871552)  # fg 1 eff rad
values.append(1.39390583)  # fg 1 sersic inex
values.append(0.83443597)  # fg 1 axis ratio
values.append(66.48069170)  # fg 1 phi
values.append(0.03447566)  # fg 2 intensity
values.append(4.52134788)  # fg 2 effective radius
values.append(0.61310712)  # fg 2 axis ratio
values.append(66.48069170)  # fg 2 phi
values.append(0.04173277)  # Lens Kappa s
values.append(2.71727396)  # Lens MLR 1
values.append(5.67223821)  # Lens MLR 2

total_mass, stellar_mass, bulge_mass, envelope_mass, dark_mass = slacs.masses_of_all_samples_ltm2(
    radius_kpc=slacs.einstein_radius * slacs.kpc_per_arcsec,
    values=values,
    center_skip=center_skip,
    ltm_skip=ltm_skip,
)

print("Masses as R_Ein (kpc) = ", slacs.einstein_radius * slacs.kpc_per_arcsec)
print()
print("SIE Mass = {:.3E}".format(sie_mass_ein_r))
print("Total Decomposed Mass = {:.3E}".format(total_mass))
print("Stellar Mass = {:.3E}".format(stellar_mass))
print("Dark Mass = {:.3E}".format(dark_mass))
print("Bulge Mass = {:.3E}".format(bulge_mass))
print("Envelope Mass = {:.3E}".format(envelope_mass))

total_mass, stellar_mass, bulge_mass, envelope_mass, dark_mass = slacs.masses_of_all_samples_ltm2(
    radius_kpc=4.75, values=values, center_skip=center_skip, ltm_skip=ltm_skip
)

print("Masses as 4.75 kpc")
print()
print("SIE Mass = {:.3E}".format(sie_mass_475))
print("Total Decomposed Mass = {:.3E}".format(total_mass))
print("Stellar Mass = {:.3E}".format(stellar_mass))
print("Dark Mass = {:.3E}".format(dark_mass))
print("Bulge Mass = {:.3E}".format(bulge_mass))
print("Envelope Mass = {:.3E}".format(envelope_mass))

slacs.density_vs_radii_ltm2(
    radius_kpc=10.0, values=values, center_skip=0, ltm_skip=1, number_bins=number_bins
)
slacs.plot_density(
    image_name=image_name,
    file_name=file_name,
    title=title,
    labels=["SIE", "Sersic", "Exponential", "NFWSph"],
)



import numpy as np

from autolens.array import grids
from autolens.array import mask as msk
from autolens.data.convolution import Convolver
from autolens.array.grids import reshape_array


class LensData(object):
    def __init__(
        self,
        ccd_data,
        mask,
        sub_grid_size=2,
        positions=None,
        positions_threshold=None,
        trimmed_psf_shape=None,
        pixel_scale_interpolation_grid=None,
        pixel_scale_binned_grid=None,
        inversion_pixel_limit=None,
        inversion_uses_border=True,
        hyper_noise_map_max=None,
        preload_pixelization_grids_of_planes=None,
    ):
        """
        The lens data is the collection of instrument (image, noise-map, PSF), a mask, grid, convolver \
        and other utilities that are used for modeling and fitting an image of a strong lens.

        Whilst the image, noise-map, etc. are loaded in 2D, the lens data creates reduced 1D arrays of each \
        for lensing calculations.

        Parameters
        ----------
        ccd_data: im.CCD
            The ccd instrument all in 2D (the image, noise-map, PSF, etc.)
        mask: msk.Mask
            The 2D mask that is applied to the image.
        sub_grid_size : int
            The size of the sub-grid used for each lens SubGrid. E.g. a value of 2 grid each image-pixel on a 2x2 \
            sub-grid.
        trimmed_psf_shape : (int, int)
            The shape of the PSF used for convolving model image generated using analytic light profiles. A smaller \
            shape will trim the PSF relative to the input image PSF, giving a faster analysis run-time.
        inversion_psf_shape : (int, int)
            The shape of the PSF used for convolving the inversion mapping_util matrix. A smaller \
            shape will trim the PSF relative to the input image PSF, giving a faster analysis run-time.
        positions : [[]]
            Lists of image-pixel coordinates (arc-seconds) that mappers close to one another in the source-plane(s), \
            used to speed up the non-linear sampling.
        pixel_scale_interpolation_grid : float
            If *True*, expensive to compute mass profile deflection angles will be computed on a sparse grid and \
            interpolated to the grid, sub and blurring grids.
        inversion_pixel_limit : int or None
            The maximum number of pixels that can be used by an inversion, with the limit placed primarily to speed \
            up run.
        pixel_scale_binned_cluster_grid : float or None
            If *True*, the hyper_galaxies image used to generate the cluster'grids weight map will be binned up to this higher \
            pixel scale to speed up the KMeans clustering algorithm.
        """

        ### CCD DATA + MASK ####

        self.ccd_data = ccd_data

        self.unmasked_image = ccd_data.image
        self.unmasked_noise_map = ccd_data.noise_map
        self.pixel_scale = ccd_data.pixel_scale
        self.psf = ccd_data.psf
        self.mask_1d = mask.array_1d_from_array_2d(array_2d=mask)
        self.image_1d = mask.array_1d_from_array_2d(array_2d=ccd_data.image)
        self.noise_map_1d = mask.array_1d_from_array_2d(array_2d=ccd_data.noise_map)
        self.mask_2d = mask
        self.sub_grid_size = sub_grid_size

        ### PSF TRIMMING + CONVOLVER ###

        if trimmed_psf_shape is None:
            self.trimmed_psf_shape = self.psf.shape
        else:
            self.trimmed_psf_shape = trimmed_psf_shape

        self.convolver = Convolver(
            mask=mask,
            blurring_mask=mask.blurring_mask_from_psf_shape(
                psf_shape=self.trimmed_psf_shape
            ),
            psf=self.psf.resized_scaled_array_from_array(
                new_shape=self.trimmed_psf_shape
            ),
        )

        ### GRIDS ###

        self.grid = grids.Grid.from_mask_and_sub_grid_size(
            mask=mask, sub_grid_size=sub_grid_size
        )

        self.pixel_scale_binned_grid = pixel_scale_binned_grid

        if pixel_scale_binned_grid is not None:
            binned_grid = grids.BinnedGrid.from_mask_and_pixel_scale_binned_grid(
                mask=mask, pixel_scale_binned_grid=pixel_scale_binned_grid
            )
            self.grid.new_grid_with_binned_grid(binned_grid=binned_grid)

        self.pixel_scale_interpolation_grid = pixel_scale_interpolation_grid

        if pixel_scale_interpolation_grid is not None:

            self.grid = self.grid.new_grid_with_interpolator(
                pixel_scale_interpolation_grid=pixel_scale_interpolation_grid
            )

        ### POSITIONS ###

        if positions is not None:
            self.positions = list(
                map(lambda position_set: np.asarray(position_set), positions)
            )
        else:
            self.positions = positions

        self.positions_threshold = positions_threshold

        self.hyper_noise_map_max = hyper_noise_map_max

        self.inversion_pixel_limit = inversion_pixel_limit
        self.inversion_uses_border = inversion_uses_border

        self.preload_blurring_grid = grids.Grid.blurring_grid_from_mask_and_psf_shape(
            mask=mask, psf_shape=self.trimmed_psf_shape
        )

        if pixel_scale_interpolation_grid is not None:

            self.preload_blurring_grid = self.preload_blurring_grid.new_grid_with_interpolator(
                pixel_scale_interpolation_grid=pixel_scale_interpolation_grid
            )

        self.preload_pixelization_grids_of_planes = preload_pixelization_grids_of_planes

    def new_lens_data_with_modified_image(self, modified_image):

        ccd_data_with_modified_image = self.ccd_data.new_ccd_data_with_modified_image(
            modified_image=modified_image
        )

        return LensData(
            ccd_data=ccd_data_with_modified_image,
            mask=self.mask_2d,
            sub_grid_size=self.sub_grid_size,
            positions=self.positions,
            positions_threshold=self.positions_threshold,
            trimmed_psf_shape=self.trimmed_psf_shape,
            pixel_scale_interpolation_grid=self.pixel_scale_interpolation_grid,
            pixel_scale_binned_grid=self.pixel_scale_binned_grid,
            inversion_pixel_limit=self.inversion_pixel_limit,
            inversion_uses_border=self.inversion_uses_border,
            hyper_noise_map_max=self.hyper_noise_map_max,
            preload_pixelization_grids_of_planes=self.preload_pixelization_grids_of_planes,
        )

    def new_lens_data_with_binned_up_ccd_data_and_mask(self, bin_up_factor):

        binned_up_ccd_data = self.ccd_data.new_ccd_data_with_binned_up_arrays(
            bin_up_factor=bin_up_factor
        )
        binned_up_mask = self.mask_2d.binned_up_mask_from_mask(
            bin_up_factor=bin_up_factor
        )

        return LensData(
            ccd_data=binned_up_ccd_data,
            mask=binned_up_mask,
            sub_grid_size=self.sub_grid_size,
            positions=self.positions,
            positions_threshold=self.positions_threshold,
            trimmed_psf_shape=self.trimmed_psf_shape,
            pixel_scale_interpolation_grid=self.pixel_scale_interpolation_grid,
            pixel_scale_binned_grid=self.pixel_scale_binned_grid,
            inversion_pixel_limit=self.inversion_pixel_limit,
            inversion_uses_border=self.inversion_uses_border,
            hyper_noise_map_max=self.hyper_noise_map_max,
            preload_pixelization_grids_of_planes=self.preload_pixelization_grids_of_planes,
        )

    def new_lens_data_with_signal_to_noise_limit(self, signal_to_noise_limit):

        ccd_data_with_signal_to_noise_limit = self.ccd_data.new_ccd_data_with_signal_to_noise_limit(
            signal_to_noise_limit=signal_to_noise_limit
        )

        return LensData(
            ccd_data=ccd_data_with_signal_to_noise_limit,
            mask=self.mask_2d,
            sub_grid_size=self.sub_grid_size,
            positions=self.positions,
            positions_threshold=self.positions_threshold,
            trimmed_psf_shape=self.trimmed_psf_shape,
            pixel_scale_interpolation_grid=self.pixel_scale_interpolation_grid,
            pixel_scale_binned_grid=self.pixel_scale_binned_grid,
            inversion_pixel_limit=self.inversion_pixel_limit,
            inversion_uses_border=self.inversion_uses_border,
            hyper_noise_map_max=self.hyper_noise_map_max,
            preload_pixelization_grids_of_planes=self.preload_pixelization_grids_of_planes,
        )

    def mask(self, return_in_2d=True):
        if return_in_2d:
            return self.mask_2d
        else:
            return self.mask_1d

    @reshape_array
    def image(self, return_in_2d=True):
        return self.image_1d

    @reshape_array
    def noise_map(self, return_in_2d=True):
        return self.noise_map_1d

    @property
    def signal_to_noise_map_1d(self):
        return self.image_1d / self.noise_map_1d

    @reshape_array
    def signal_to_noise_map(self, return_in_2d=True):
        return self.signal_to_noise_map_1d

    def __array_finalize__(self, obj):
        if isinstance(obj, LensData):
            self.ccd_data = obj.ccd_data
            self.unmasked_image = obj.unmasked_image
            self.unmasked_noise_map = obj.unmasked_noise_map
            self.mask_2d = obj.mask_2d
            self.mask_1d = obj.mask_1d
            self.psf = obj.psf
            self.trimmed_psf_shape = obj.trimmed_psf_shape
            self.mask_1d = obj.mask_1d
            self.image_1d = obj.image_1d
            self.noise_map_1d = obj.noise_map_1d
            self.sub_grid_size = obj.sub_grid_size
            self.convolver = obj.convolver
            self.grid = obj.grid
            self.positions = obj.positions
            self.pixel_scale_interpolation_grid = obj.pixel_scale_interpolation_grid
            self.pixel_scale_binned_grid = obj.pixel_scale_binned_grid
            self.inversion_uses_border = obj.inversion_uses_border
            self.inversion_pixel_limit = obj.inversion_pixel_limit
            self.hyper_noise_map_max = obj.hyper_noise_map_max
            self.preload_blurring_grid = obj.preload_blurring_grid
            self.preload_pixelization_grids_of_planes = (
                obj.preload_pixelization_grids_of_planes
            )



import autofit as af
import autolens as al
from autolens.pipeline import pipeline

# All pipelines begin with a comment describing the pipeline and a phase-by-phase description of what it does.

# In this pipeline, we'll perform a basic analysis which fits a source galaxy using a parametric light profile and a
# lens galaxy where its light is included and fitted, using three phases:

# Phase 1) Fit the lens galaxy's light using an elliptical Sersic light profile.

# Phase 2) Use this lens subtracted image to fit the lens galaxy's mass (SIE) and source galaxy's light (Sersic).

# Phase 3) Fit the lens's light, mass and source's light simultaneously using priors initialized from the above 2 phases.


def make_pipeline(phase_folders=None):

    # Pipelines takes 'phase_folders' as input, which in conjunction with the pipeline name specify the path structure of
    # the output. In the pipeline runner we pass the phase_folders ['howtolens, c3_t1_lens_and_source], which means the
    # output of this pipeline go to the folder 'autolens_workspace/output/howtolens/c3_t1_lens_and_source/pipeline__light_and_source'.

    # By default, the pipeline folders is None, meaning the output go to the directory 'output/pipeline_name',
    # which in this case would be 'output/pipeline_light_and_source'.

    # In the example pipelines found in 'autolens_workspace/pipelines' folder, we pass the name of our strong lens data
    # to the pipeline path. This allows us to fit a large sample of lenses using one pipeline and store all of their
    # results in an ordered directory structure.

    pipeline_name = "pipeline__light_and_source"

    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/phase_folder_1/phase_folder_2/pipeline_name/pipeline_tag/phase_name/phase_tag/'
    phase_folders.append(pipeline_name)

    ### PHASE 1 ###

    # In chapter 2, we learnt how to mask data. With pipelines we can change our mask between phases. Afterall,
    # the bigger the mask, the slower the run-time. In the early phases of a pipeline we're not bothered about fitting
    # all of the image. Aggresive masking (which removes lots of image-pixels) is an appealing way to get things running fast.

    # In phase 1, we're only interested in fitting the lens's light, so we'll mask out the source-galaxy entirely.
    # This'll give us a nice speed up and ensure the source's light doesn't impact our light-profile fit.

    # We want a mask that is shaped like the source-galaxy. The shape of the source is an 'annulus' (e.g. a ring),
    # so we're going to use an annular mask. For example, if an annulus is specified between an inner radius of 0.5"
    # and outer radius of 2.0", all pixels in two rings between 0.5" and 2.0" are included in the analysis.

    # But wait, we actually want the opposite of this! We want a masks where the pixels between 0.5" and 2.0" are not
    # included! They're the pixels the source is actually located. Therefore, we're going to use an 'anti-annular
    # mask', where the inner and outer radii are the regions we omit from the analysis. This means we need to specify
    # a third mask radii, further out, such that data at these exterior edges of the image are masked.

    # We can set a mask using a 'mask_function', which returns the mask used by a phase.

    def mask_function(image):
        return al.Mask.circular_anti_annular(
            shape=image.shape,
            pixel_scale=image.pixel_scale,
            inner_radius_arcsec=0.5,
            outer_radius_arcsec=1.6,
            outer_radius_2_arcsec=2.5,
        )

    # Next , wecreate the phase, using the same notation we learnt before (noting the masks function is passed to
    # this phase ensuring the anti-annular masks above is used).

    phase1 = al.PhaseImaging(
        phase_name="phase_1__lens_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(redshift=0.5, light=al.light_profiles.EllipticalSersic)
        ),
        mask_function=mask_function,
        optimizer_class=af.MultiNest,
    )

    # We'll use the MultiNest black magic we covered in tutorial 7 of chapter 2 to get this phase to run fast.

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 30
    phase1.optimizer.sampling_efficiency = 0.5

    ### PHASE 2 ###

    # In phase 2, we fit the source galaxy's light. Thus, we want to make 2 changes from the previous phase.

    # 1) We want to fit the lens subtracted image calculated in phase 1, instead of the observed image.
    # 2) We want to mask the central regions of this image where there are residuals due to the lens light subtraction.

    # We can use the mask function again, to modify the mask to an annulus. We'll use the same ring radii as before.

    def mask_function(image):
        return al.Mask.circular_annular(
            shape=image.shape,
            pixel_scale=image.pixel_scale,
            inner_radius_arcsec=0.5,
            outer_radius_arcsec=3.0,
        )

    # To modify an image, we call a new function, 'modify image'. This function behaves like the pass-priors functions
    # before, whereby we create a python 'class' in a Phase to set it up.  This ensures it has access to the pipeline's
    # 'results' (which you may have noticed was in the the customize_priors functions as well).

    # To setup the modified image we take the observed image and subtract-off the model image from the
    # previous phase, which, if you're keeping track, is an image of the lens galaxy. However, if we just used the
    # 'model_image' in the fit, this would only include pixels that were masked. We want to subtract the lens off the
    # entire image - fortunately, PyAutoLens automatically generates an 'unmasked_lens_plane_model_image' as well!

    class LensSubtractedPhase(al.PhaseImaging):
        def modify_image(self, image, results):
            phase_1_results = results.from_phase("phase_1__lens_sersic")
            return image - phase_1_results.unmasked_model_image_of_planes[0]

    # The function above demonstrates the most important thing about pipelines - that every phase has access to the
    # results of all previous phases. This means we can feed information through the pipeline and therefore use the
    # results of previous phases to setup new phases.

    # You should see that this is done by using the phase_name of the phase we're interested in, which in the above
    # code is named 'phase_1__lens_sersic' (you can check this on line 73 above).

    # We'll do this again in phase 3 and throughout all of the pipelines in this chapter and the workspace examples.

    # We setup phase 2 as per usual. Note that we don't need to pass the modify image function.

    phase2 = LensSubtractedPhase(
        phase_name="phase_2__lens_sie__source_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5, mass=al.mass_profiles.EllipticalIsothermal
            ),
            source=al.GalaxyModel(
                redshift=1.0, light=al.light_profiles.EllipticalSersic
            ),
        ),
        mask_function=mask_function,
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 50
    phase2.optimizer.sampling_efficiency = 0.3

    ### PHASE 3 ###

    # Finally, in phase 3, we want to fit the lens and source simultaneously.

    # We'll use the 'customize_priors' function that we all know and love to do this. However, we're going to use the
    # 'results' argument that, in chapter 2, we ignored. This stores the results of the lens model of
    # phases 1 and 2 meaning we can use it to initialize phase 3's priors!

    class LensSourcePhase(al.PhaseImaging):
        def customize_priors(self, results):

            # The previous results is a 'list' in python. The zeroth index entry of the list maps to the results of
            # phase 1, the first entry to phase 2, and so on.

            phase_1_results = results.from_phase("phase_1__lens_sersic")
            phase_2_results = results.from_phase("phase_2__lens_sie__source_sersic")

            # To link two priors together we invoke the 'variable' attribute of the previous results. By invoking
            # 'variable', this means that:

            # 1) The parameter will be a free-parameter fitted for by the non-linear search.
            # 2) It will use a GaussianPrior based on the previous results as its initialization (we'll cover how this
            #    Gaussian is setup in tutorial 4, for now just imagine it links the results in a sensible way).

            # We can simply link every source galaxy parameter to its phase 2 inferred value, as follows

            self.galaxies.source.light.centre_0 = (
                phase_2_results.variable.galaxies.source.light.centre_0
            )

            self.galaxies.source.light.centre_1 = (
                phase_2_results.variable.galaxies.source.light.centre_1
            )

            self.galaxies.source.light.axis_ratio = (
                phase_2_results.variable.galaxies.source.light.axis_ratio
            )

            self.galaxies.source.light.phi = (
                phase_2_results.variable.galaxies.source.light.phi
            )

            self.galaxies.source.light.intensity = (
                phase_2_results.variable.galaxies.source.light.intensity
            )

            self.galaxies.source.light.effective_radius = (
                phase_2_results.variable.galaxies.source.light.effective_radius
            )

            self.galaxies.source.light.sersic_index = (
                phase_2_results.variable.galaxies.source.light.sersic_index
            )

            # However, listing every parameter like this is ugly and becomes cumbersome if we have a lot of parameters.

            # If, like in the above example, you are making all of the parameters of a lens or source galaxy variable,
            # you can simply set the source galaxy equal to one another without specifying each parameter of every
            # light and mass profile.

            self.galaxies.source = (
                phase_2_results.variable.galaxies.source
            )  # This is identical to lines 196-203 above.

            # For the lens galaxies we have a slightly weird circumstance where the light profiles requires the
            # results of phase 1 and the mass profile the results of phase 2. When passing these as a 'variable', we
            # can split them as follows

            self.galaxies.lens.light = phase_1_results.variable.galaxies.lens.light

            self.galaxies.lens.mass = phase_2_results.variable.galaxies.lens.mass

    phase3 = LensSourcePhase(
        phase_name="phase_3__lens_sersic_sie__source_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                light=al.light_profiles.EllipticalSersic,
                mass=al.mass_profiles.EllipticalIsothermal,
            ),
            source=al.GalaxyModel(
                redshift=1.0, light=al.light_profiles.EllipticalSersic
            ),
        ),
        optimizer_class=af.MultiNest,
    )

    phase3.optimizer.const_efficiency_mode = True
    phase3.optimizer.n_live_points = 50
    phase3.optimizer.sampling_efficiency = 0.3

    return pipeline.PipelineImaging(pipeline_name, phase1, phase2, phase3)



import time

from autolens.model.profiles import mass_profiles as mp
from autolens.lens import lens_data as ld
from autolens.array import mask as msk

from test.simulation import simulation_util

# Although we could test the deflection angles without using an image (e.al. by just making a grid), we have chosen to
# set this test up using an image and mask. This gives run-time numbers that can be easily related to an actual lens
# analysis

sub_grid_size = 2
radius_arcsec = 4.0

print("sub grid size = " + str(sub_grid_size))
print("circular mask radius = " + str(radius_arcsec) + "\n")

for data_resolution in ["LSST", "Euclid", "HST", "HST_Up", "AO"]:

    ccd_data = simulation_util.load_test_ccd_data(
        data_type="lens_mass__source_smooth",
        data_resolution=data_resolution,
        psf_shape=(3, 3),
    )
    mask = al.Mask.circular(
        shape=ccd_data.shape,
        pixel_scale=ccd_data.pixel_scale,
        radius_arcsec=radius_arcsec,
    )
    lens_data = al.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=sub_grid_size)

    print("Deflection angle run times for image type " + data_resolution + "\n")
    print("Number of points = " + str(lens_data.grid.shape[0]) + "\n")

    ### EllipticalIsothermal ###

    mass_profile = al.EllipticalIsothermal(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, einstein_radius=1.0
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("EllipticalIsothermal time = {}".format(diff))

    ### SphericalIsothermal ###

    mass_profile = al.mass_profiles.SphericalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.0
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("SphericalIsothermal time = {}".format(diff))

    ### EllipticalPowerLaw (slope = 1.5) ###

    mass_profile = al.EllipticalPowerLaw(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, einstein_radius=1.0, slope=1.5
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("EllipticalPowerLaw (slope = 1.5) time = {}".format(diff))

    ### SphericalPowerLaw (slope = 1.5) ###

    mass_profile = al.SphericalPowerLaw(
        centre=(0.0, 0.0), einstein_radius=1.0, slope=1.5
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("SphericalPowerLaw (slope = 1.5) time = {}".format(diff))

    ### EllipticalPowerLaw (slope = 2.5) ###

    mass_profile = al.EllipticalPowerLaw(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, einstein_radius=1.0, slope=2.5
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("EllipticalPowerLaw (slope = 2.5) time = {}".format(diff))

    ### SphericalPowerLaw (slope = 2.5) ###

    mass_profile = al.SphericalPowerLaw(
        centre=(0.0, 0.0), einstein_radius=1.0, slope=2.5
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("SphericalPowerLaw (slope = 2.5) time = {}".format(diff))

    ### EllipticalCoredPowerLaw ###

    mass_profile = al.EllipticalCoredPowerLaw(
        centre=(0.0, 0.0),
        axis_ratio=0.8,
        phi=45.0,
        einstein_radius=1.0,
        slope=2.0,
        core_radius=0.1,
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("EllipticalCoredPowerLaw time = {}".format(diff))

    ### SphericalCoredPowerLaw ###

    mass_profile = al.SphericalCoredPowerLaw(
        centre=(0.0, 0.0), einstein_radius=1.0, slope=2.0, core_radius=0.1
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("SphericalCoredPowerLaw time = {}".format(diff))

    ### EllipticalGeneralizedNFW (inner_slope = 0.5) ###

    # mass_profile = al.EllipticalGeneralizedNFW(centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, kappa_s=0.1,
    #                                            scale_radius=10.0, inner_slope=0.5)
    #
    # start = time.time()
    # mass_profile.deflections_from_grid(grid=lens_data.grid)
    # diff = time.time() - start
    # print("EllipticalGeneralizedNFW (inner_slope = 1.0) time = {}".format(diff))

    ### SphericalGeneralizedNFW (inner_slope = 0.5) ###

    mass_profile = al.SphericalGeneralizedNFW(
        centre=(0.0, 0.0), kappa_s=0.1, scale_radius=10.0, inner_slope=0.5
    )

    # start = time.time()
    # mass_profile.deflections_from_grid(grid=lens_data.grid)
    # diff = time.time() - start
    # print("SphericalGeneralizedNFW (inner_slope = 1.0) time = {}".format(diff))

    ### EllipticalNFW ###

    mass_profile = al.EllipticalNFW(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, kappa_s=0.1, scale_radius=10.0
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("EllipticalNFW time = {}".format(diff))

    ### SphericalNFW ###

    mass_profile = al.SphericalNFW(centre=(0.0, 0.0), kappa_s=0.1, scale_radius=10.0)

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("SphericalNFW time = {}".format(diff))

    ### SphericalNFW ###

    mass_profile = al.SphericalTruncatedNFW(
        centre=(0.0, 0.0), kappa_s=0.1, scale_radius=10.0, truncation_radius=5.0
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("SphericalTruncatedNFW time = {}".format(diff))

    ### EllipticalExponential ###

    profile = al.EllipticalExponential(
        centre=(0.0, 0.0),
        axis_ratio=0.8,
        phi=45.0,
        intensity=1.0,
        effective_radius=1.0,
        mass_to_light_ratio=1.0,
    )

    start = time.time()
    profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("EllipticalExponential time = {}".format(diff))

    ### SphericalExponential ###

    profile = al.SphericalExponential(
        centre=(0.0, 0.0), intensity=1.0, effective_radius=1.0, mass_to_light_ratio=1.0
    )

    start = time.time()
    profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("SphericalExponential time = {}".format(diff))

    ### EllipticalDevVaucouleurs ###

    profile = al.EllipticalDevVaucouleurs(
        centre=(0.0, 0.0),
        axis_ratio=0.8,
        phi=45.0,
        intensity=1.0,
        effective_radius=1.0,
        mass_to_light_ratio=1.0,
    )

    start = time.time()
    profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("EllipticalDevVaucouleurs time = {}".format(diff))

    ### SphericalDevVaucouleurs ###

    profile = al.SphericalDevVaucouleurs(
        centre=(0.0, 0.0), intensity=1.0, effective_radius=1.0, mass_to_light_ratio=1.0
    )

    start = time.time()
    profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("SphericalDevVaucouleurs time = {}".format(diff))

    ### EllipticalSersic ###

    mass_profile = al.EllipticalSersic(
        centre=(0.0, 0.0),
        axis_ratio=0.8,
        phi=45.0,
        intensity=1.0,
        effective_radius=1.0,
        sersic_index=2.5,
        mass_to_light_ratio=1.0,
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("EllipticalSersic time = {}".format(diff))

    ### SphericalSersic ###

    mass_profile = al.SphericalSersic(
        centre=(0.0, 0.0),
        intensity=1.0,
        effective_radius=1.0,
        sersic_index=2.5,
        mass_to_light_ratio=1.0,
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("SphericalSersic time = {}".format(diff))

    ### EllipticalSersicRadialGradient (gradient = -1.0) ###

    mass_profile = al.EllipticalSersicRadialGradient(
        centre=(0.0, 0.0),
        axis_ratio=0.8,
        phi=45.0,
        intensity=1.0,
        effective_radius=1.0,
        sersic_index=2.5,
        mass_to_light_ratio=1.0,
        mass_to_light_gradient=-1.0,
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("EllipticalSersicRadialGradient (gradient = -1.0) time = {}".format(diff))

    ### SphericalersicRadialGradient (gradient = -1.0) ###

    mass_profile = al.SphericalSersicRadialGradient(
        centre=(0.0, 0.0),
        intensity=1.0,
        effective_radius=1.0,
        sersic_index=2.5,
        mass_to_light_ratio=1.0,
        mass_to_light_gradient=-1.0,
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("SphericalSersicRadialGradient (gradient = -1.0) time = {}".format(diff))

    ### EllipticalSersicRadialGradient (gradient = 1.0) ###

    mass_profile = al.EllipticalSersicRadialGradient(
        centre=(0.0, 0.0),
        axis_ratio=0.8,
        phi=45.0,
        intensity=1.0,
        effective_radius=1.0,
        sersic_index=2.5,
        mass_to_light_ratio=1.0,
        mass_to_light_gradient=1.0,
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("EllipticalSersicRadialGradient (gradient = 1.0) time = {}".format(diff))

    ### SphericalersicRadialGradient (gradient = 1.0) ###

    mass_profile = al.SphericalSersicRadialGradient(
        centre=(0.0, 0.0),
        intensity=1.0,
        effective_radius=1.0,
        sersic_index=2.5,
        mass_to_light_ratio=1.0,
        mass_to_light_gradient=1.0,
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("SphericalSersicRadialGradient (gradient = 1.0) time = {}".format(diff))

    print()



import autolens as al
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg
from autolens.model.inversion.plotters import inversion_plotters
from autolens.model.inversion.plotters import mapper_plotters

# In the previous tutorial, we told our invrsion to use a border. Here, we'll discuss what this border does.

# To begin, lets simulate a simple image and use it to generate a rectangular mapper, as we're now used to doing.
def simulate():

    psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
        shape=(180, 180), pixel_scale=0.05
    )

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0, einstein_radius=1.6
        ),
    )

    source_galaxy = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.1, 0.1),
            axis_ratio=0.8,
            phi=90.0,
            intensity=0.2,
            effective_radius=0.3,
            sersic_index=1.0,
        ),
    )

    tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

    return al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
        tracer=tracer,
        grid=grid,
        pixel_scale=0.05,
        exposure_time=300.0,
        psf=psf,
        background_sky_level=0.5,
        add_noise=True,
    )


# Lets have a quick look at the image.
ccd_data = simulate()
al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

# So, what is a border? In the image-plane, a border is the set of exterior pixels in a mask that are at, well, its
# border. Lets plot the image with a circular mask, and tell our imaging plotter to plot the border as well.
mask_circular = al.Mask.circular(
    shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=2.5
)

al.ccd_plotters.plot_ccd_subplot(
    ccd_data=ccd_data, mask=mask_circular, should_plot_border=True
)

# As you can see, for a circular mask, the border *is* the edge of our mask (the ring of black dots we're used to seeing
# whenever we plot a mask). For an annular mask, not every pixel on the edge of the mask is necessarily a part of its
# border!
mask_annular = al.Mask.circular_annular(
    shape=ccd_data.shape,
    pixel_scale=ccd_data.pixel_scale,
    inner_radius_arcsec=0.8,
    outer_radius_arcsec=2.5,
)

al.ccd_plotters.plot_ccd_subplot(
    ccd_data=ccd_data, mask=mask_annular, should_plot_border=True
)

# Indeed, a border is *only* the pixels at the exterior edge of our mask, which for the annular mask above means non of
# the pixels at the inner radius = 0.8" edge are part of the border.

# So, what does a border actually do? To show you, we'll need to fit this image with a lens model and mapper and we'll
# do that by using the same function as the previous tutorial (to perform a quick source galaxy fit) but with the
# option to input a mask and use a border.
def perform_fit_with_source_galaxy_mask_and_border(
    source_galaxy, mask, inversion_uses_border
):

    ccd_data = simulate()

    lens_data = al.LensData(
        ccd_data=ccd_data,
        mask=mask,
        sub_grid_size=2,
        inversion_uses_border=inversion_uses_border,
    )

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0, einstein_radius=1.6
        ),
    )

    tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

    return al.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)


# Okay, so lets first look at our mapper without using a border using our annular mask.
source_galaxy = al.Galaxy(
    redshift=1.0,
    pixelization=pix.Rectangular(shape=(40, 40)),
    regularization=reg.Constant(coefficient=1.0),
)

fit = perform_fit_with_source_galaxy_mask_and_border(
    source_galaxy=source_galaxy, mask=mask_annular, inversion_uses_border=False
)

inversion_plotters.plot_pixelization_values(
    inversion=fit.inversion, should_plot_grid=True
)

# Everything looks fine - we get a reconstructed source on a visually appeasing source-plane grid. So, why are we
# so worried about borders? Lets see what happens if we use a circular mask instead.
fit = perform_fit_with_source_galaxy_mask_and_border(
    source_galaxy=source_galaxy, mask=mask_circular, inversion_uses_border=False
)

inversion_plotters.plot_pixelization_values(
    inversion=fit.inversion, should_plot_grid=True
)

# Woah - whats happened? There are lots of extra points on our source-plane grid which trace to extremely large radii
# away from the central regions of the source-plane! These points are traced image-pixels (just like all the other
# points) which correspond to the central image-pixels that our annular mask masked but that our circular mask didn't!

# Lets quickly check this using a mapper plotter.
mapper_plotters.plot_image_and_mapper(
    ccd_data=ccd_data,
    mapper=fit.inversion.mapper,
    mask=mask_circular,
    should_plot_grid=True,
    image_pixels=[
        [range(3765, 3795)],
        [range(4065, 4095)],
        [range(3865, 3895)],
        [range(3965, 3995)],
        [range(4165, 4195)],
    ],
)

# So, whats happening physically? Towards the centre of our EllipticalIsothermal mass profile the density profile
# becomes extremely cuspy (rising very sharply). This cause extremely large deflection angles to be computed - lets
# have a quick look.
al.ray_tracing_plotters.plot_deflections_y(tracer=fit.tracer, grid=fit.grid)
al.ray_tracing_plotters.plot_deflections_x(tracer=fit.tracer, grid=fit.grid)

# This means that our central image pixels are highly demagnified, tracing to extremely large values in the source
# plane! Physically, this isn't a problem - it just means that we don't see a 'central image' in most strong lenses
# as light-rays which trace through the centre of the lens are demagnified. However, if the lens galaxy had a cored mass
# distribution we would see the central image.

# This is a problem for our pixelization and mapper, which in the source-plane fits these demagnified pixels
# like any other pixels. This has two negative consequences:

# 1) The rectangular grid we 'overlay' over the source-plane is much larger than for the annular mask because it has to
#    expand to include the demagnified image-pixels. As a result, large source-pixels are used to reconstruct the
#    central regions of the source-plane (where the source galaxy is actually located), meaning we reconstruct the
#    source-galaxy at a lower effective resolution.

# 2) The rectangular grid reconstructs the flux of the demanigified image pixels using source-pixels which contain
#    *only* demagnified image pixels. However, these source-pixels *should* have other image-pixels traced within
#    them from pixels at large radii from the centre of the lens galaxy. Unfortunately, our circular mask
#    masks these pixels out, meaning they do not make it to our source-plane and are omitted from the source reconstruction.

# Lets quickly use a larger circular mask to confirm that these pixels do exist, if we don't mask them.
mask_circular_large = al.Mask.circular(
    shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=4.0
)

fit = perform_fit_with_source_galaxy_mask_and_border(
    source_galaxy=source_galaxy, mask=mask_circular, inversion_uses_border=False
)

inversion_plotters.plot_pixelization_values(
    inversion=fit.inversion, should_plot_grid=True
)

# This second point is a *huge* problem, as allowing source-pixels to fit regions of our mask in this completely
# unphysical way introduces extremely dangerous systematics into our source reconstruction and lens model analysis.
# You can see this in the weird patterns these pixels make in the exterior regions of our source-reconstruction!

# Borders are the solution to this problem. We simply take the mask border in the image-plane we showed above,
# trace it to the source-plane and relocate all traced image-pixels pixels outside this source-plane border to its
# edge. Lets take a look.
fit = perform_fit_with_source_galaxy_mask_and_border(
    source_galaxy=source_galaxy, mask=mask_circular, inversion_uses_border=True
)

inversion_plotters.plot_pixelization_values(
    inversion=fit.inversion, should_plot_grid=True
)

mapper_plotters.plot_image_and_mapper(
    ccd_data=ccd_data,
    mapper=fit.inversion.mapper,
    mask=mask_circular,
    should_plot_grid=True,
    image_pixels=[
        [range(3765, 3795)],
        [range(4065, 4095)],
        [range(3865, 3895)],
        [range(3965, 3995)],
        [range(4165, 4195)],
    ],
)

# This successfully addresses both of the issues above! However, you might be thinking, isn't that a bit of a hack? Its
# not really a physical treatment of the ray-tracing, is it?

# Well, you're right. However, the *only* physical way to do this would be to use a mask so large that all demangified
# central pixels are surrounded by traced image-pixels. This would require a mask so large our computer would crash,
# That's not a good solution, thus borders provide us with a workaround - one that I've extensively tested and have
# found that, provided your mask isn't too small, doesn't lead to systematic biases.

# Next, I'm going to quickly highlight how important borders are when modeling multiple lens galaxies. Their complex
# mass distribution and lensing configuration often produce very nasty edge effects where image pixels not just in the
# centre of mask, but anywhere in the mask, trace beyond the source-plane border.
def simulate_image_x2_lenses():

    psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
        shape=(300, 300), pixel_scale=0.05
    )

    lens_galaxy_0 = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(1.1, 0.51), axis_ratio=0.9, phi=110.0, einstein_radius=1.07
        ),
    )

    lens_galaxy_1 = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(-0.20, -0.35), axis_ratio=0.56, phi=16.0, einstein_radius=0.71
        ),
    )

    source_galaxy_0 = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.05, 0.05),
            axis_ratio=0.8,
            phi=90.0,
            intensity=0.2,
            effective_radius=0.3,
            sersic_index=1.0,
        ),
    )

    tracer = al.Tracer.from_galaxies(
        galaxies=[lens_galaxy_0, lens_galaxy_1, source_galaxy_0]
    )

    return al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
        tracer=tracer,
        grid=grid,
        pixel_scale=0.05,
        exposure_time=300.0,
        psf=psf,
        background_sky_level=0.1,
        add_noise=True,
    )


# Lets simulate our 2 lens system, define a new circular mask and plot them.
ccd_data = simulate_image_x2_lenses()

mask_circular = al.Mask.circular(
    shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=2.8
)

al.ccd_plotters.plot_ccd_subplot(
    ccd_data=ccd_data,
    mask=mask_circular,
    extract_array_from_mask=True,
    zoom_around_mask=True,
    should_plot_border=True,
)


# We need to redefine our perform fit function, to use the x2 lens galaxy model.
def perform_fit_x2_lenses_with_source_galaxy_mask_and_border(
    source_galaxy, mask, inversion_uses_border
):

    simulate_image_x2_lenses()

    lens_data = al.LensData(
        ccd_data=ccd_data, mask=mask, inversion_uses_border=inversion_uses_border
    )

    lens_galaxy_0 = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(1.1, 0.51), axis_ratio=0.9, phi=110.0, einstein_radius=1.07
        ),
    )

    lens_galaxy_1 = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(-0.20, -0.35), axis_ratio=0.56, phi=16.0, einstein_radius=0.71
        ),
    )

    tracer = al.Tracer.from_galaxies(
        galaxies=[lens_galaxy_0, lens_galaxy_1, source_galaxy]
    )

    return al.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)


# Now, lets fit this image using the input model and perform the source reconstruction without a border. As you can see,
# we get many demagnified image pixels which trace well beyond our source-plane border if we don't relocate them!
fit = perform_fit_x2_lenses_with_source_galaxy_mask_and_border(
    source_galaxy=source_galaxy, mask=mask_circular, inversion_uses_border=False
)

inversion_plotters.plot_pixelization_values(
    inversion=fit.inversion, should_plot_grid=True, should_plot_border=True
)

# However, when we relocate them, we get a good-looking source-plane with a well defined border and edge, thus ensuring
# our analysis will (hopefully) be free of systematic biases.
fit = perform_fit_x2_lenses_with_source_galaxy_mask_and_border(
    source_galaxy=source_galaxy, mask=mask_circular, inversion_uses_border=True
)

inversion_plotters.plot_pixelization_values(
    inversion=fit.inversion, should_plot_grid=True, should_plot_border=True
)

# Multi-galaxy modeling is rife for border effects and if you have multiple lens galaxies I heartily recommend you
# # pay a close eye to your source-plane borders!
#
# # Before we end,I want to quickly highlight that care must be taken when choosing the size of your mask. If you don't
# # choose a big enough mask, the border won't be able to relocate all of the demanigified image pixels to the border
# # edge.
mask_circular = al.Mask.circular(
    shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=2.5
)

fit = perform_fit_x2_lenses_with_source_galaxy_mask_and_border(
    source_galaxy=source_galaxy, mask=mask_circular, inversion_uses_border=True
)

inversion_plotters.plot_pixelization_values(
    inversion=fit.inversion, should_plot_grid=True, should_plot_border=True
)


mask_circular = al.Mask.circular(
    shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=2.7
)

fit = perform_fit_x2_lenses_with_source_galaxy_mask_and_border(
    source_galaxy=source_galaxy, mask=mask_circular, inversion_uses_border=True
)

inversion_plotters.plot_pixelization_values(
    inversion=fit.inversion, should_plot_grid=True, should_plot_border=True
)


mask_circular = al.Mask.circular(
    shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=2.9
)

fit = perform_fit_x2_lenses_with_source_galaxy_mask_and_border(
    source_galaxy=source_galaxy, mask=mask_circular, inversion_uses_border=True
)

inversion_plotters.plot_pixelization_values(
    inversion=fit.inversion, should_plot_grid=True, should_plot_border=True
)


mask_circular = al.Mask.circular(
    shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=3.1
)

fit = perform_fit_x2_lenses_with_source_galaxy_mask_and_border(
    source_galaxy=source_galaxy, mask=mask_circular, inversion_uses_border=True
)

inversion_plotters.plot_pixelization_values(
    inversion=fit.inversion, should_plot_grid=True, should_plot_border=True
)

# And with that, borders are done. In truth, borders should pretty much take care of themselves when you're
# using PyAutoLens and you probably won't think about them much. However, as I showed above, if you don't choose a
# large enough mask things can go wrong - thus, its important you know what borders are, so you can look out for this
# potential source of systematics!



import numpy as np

import autolens as al
from autolens.array import scaled_array
from autolens.data import fourier_transform
from autolens.data.instrument import interferometer


class MockImage(object):
    def __new__(cls, shape, value, pixel_scale=1.0):
        return al.ScaledSquarePixelArray(
            array=value * np.ones(shape=shape), pixel_scale=pixel_scale
        )


class MockNoiseMap(object):
    def __new__(cls, shape, value, pixel_scale=1.0):
        return al.ScaledSquarePixelArray(
            array=value * np.ones(shape=shape), pixel_scale=pixel_scale
        )


class MockBackgroundNoiseMap(object):
    def __new__(cls, shape, value, pixel_scale=1.0):
        return al.ScaledSquarePixelArray(
            array=value * np.ones(shape=shape), pixel_scale=pixel_scale
        )


class MockPoissonNoiseMap(object):
    def __new__(cls, shape, value, pixel_scale=1.0):
        return al.ScaledSquarePixelArray(
            array=value * np.ones(shape=shape), pixel_scale=pixel_scale
        )


class MockExposureTimeMap(object):
    def __new__(cls, shape, value, pixel_scale=1.0):
        return al.ScaledSquarePixelArray(
            array=value * np.ones(shape=shape), pixel_scale=pixel_scale
        )


class MockBackgrondSkyMap(object):
    def __new__(cls, shape, value, pixel_scale=1.0):
        return al.ScaledSquarePixelArray(
            array=value * np.ones(shape=shape), pixel_scale=pixel_scale
        )


class MockPSF(object):
    def __new__(cls, shape, value, pixel_scale=1.0, *args, **kwargs):
        return al.PSF(
            array=value * np.ones(shape=shape),
            pixel_scale=pixel_scale,
            origin=(0.0, 0.0),
        )


class MockImage1D(np.ndarray):
    def __new__(cls, shape, value, pixel_scale=1.0):
        array = value * np.ones(shape=shape)

        obj = np.array(array, dtype="float64").view(cls)
        obj.pixel_scale = pixel_scale
        obj.origin = (0.0, 0.0)

        return obj


class MockNoiseMap1D(np.ndarray):
    def __new__(cls, shape, value, pixel_scale=1.0):
        array = value * np.ones(shape=shape)

        obj = np.array(array, dtype="float64").view(cls)
        obj.pixel_scale = pixel_scale
        obj.origin = (0.0, 0.0)

        return obj


class MockCCDData(al.CCDData):
    def __init__(
        self,
        image,
        pixel_scale,
        psf,
        noise_map,
        background_noise_map,
        poisson_noise_map,
        exposure_time_map,
        background_sky_map,
        name,
    ):
        super(MockCCDData, self).__init__(
            image=image,
            pixel_scale=pixel_scale,
            psf=psf,
            noise_map=noise_map,
            background_noise_map=background_noise_map,
            poisson_noise_map=poisson_noise_map,
            exposure_time_map=exposure_time_map,
            background_sky_map=background_sky_map,
            name=name,
        )


class MockPrimaryBeam(object):
    def __new__(cls, shape, value, pixel_scale=1.0, *args, **kwargs):
        return al.PSF(
            array=value * np.ones(shape=shape),
            pixel_scale=pixel_scale,
            origin=(0.0, 0.0),
        )


class MockVisibilities(np.ndarray):
    def __new__(cls, shape, value, pixel_scale=1.0):
        array = value * np.ones(shape=(shape, 2))

        obj = np.array(array, dtype="float64").view(cls)
        obj.pixel_scale = pixel_scale
        obj.origin = (0.0, 0.0)

        return obj


class MockVisibilitiesNoiseMap(np.ndarray):
    def __new__(cls, shape, value, pixel_scale=1.0):
        array = value * np.ones(shape=shape)

        obj = np.array(array, dtype="float64").view(cls)
        obj.pixel_scale = pixel_scale
        obj.origin = (0.0, 0.0)

        return obj


class MockUVWavelengths(np.ndarray):
    def __new__(cls, shape, value, pixel_scale=1.0):
        array = np.array(
            [
                [-55636.4609375, 171376.90625],
                [-6903.21923828, 51155.578125],
                [-63488.4140625, 4141.28369141],
                [55502.828125, 47016.7265625],
                [54160.75390625, -99354.1796875],
                [-9327.66308594, -95212.90625],
            ]
        )

        obj = np.array(array, dtype="float64").view(cls)
        obj.pixel_scale = pixel_scale
        obj.origin = (0.0, 0.0)

        return obj


class MockInterferometerData(al.InterferometerData):
    def __init__(
        self,
        image,
        pixel_scale,
        psf,
        primary_beam,
        noise_map,
        visibilities,
        visibilities_noise_map,
        uv_wavelengths,
    ):
        super(MockInterferometerData, self).__init__(
            image=image,
            pixel_scale=pixel_scale,
            psf=psf,
            primary_beam=primary_beam,
            noise_map=noise_map,
            visibilities=visibilities,
            visibilities_noise_map=visibilities_noise_map,
            uv_wavelengths=uv_wavelengths,
        )


class MockTransformer(fourier_transform.Transformer):
    def __init__(self, uv_wavelengths, grid_radians):
        super(MockTransformer, self).__init__(
            uv_wavelengths=uv_wavelengths, grid_radians=grid_radians
        )



from autolens.data.instrument import abstract_data
from autolens.data.instrument import ccd
from autolens.array import mask as msk
from autolens.model.profiles import light_profiles as lp
from autolens.model.galaxy import galaxy as g
from autolens.lens import ray_tracing
from autolens.lens import lens_fit
from autolens.lens import lens_data as ld
from autolens.data.plotters import ccd_plotters
from autolens.lens.plotters import lens_fit_plotters
from test.simulation import simulation_util

# In this tutorial, we'll introduce a new pixelization, called an adaptive-pixelization. This pixelization doesn't use
# uniform grid of rectangular pixels, but instead uses ir'Voronoi' pixels. So, why would we want to do that?
# Lets take another look at the rectangular grid, and think about its weakness.

# Lets quickly remind ourselves of the image, and the 3.0" circular mask we'll use to mask it.
ccd_data = simulation_util.load_test_ccd_data(
    data_type="lens_light_dev_vaucouleurs", data_resolution="LSST"
)
mask = al.Mask.circular(
    shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=3.0
)

# The lines of code below do everything we're used to, that is, setup an image and its al.ogrid, mask it, trace it
# via a tracer, setup the rectangular mapper, etc.
lens_galaxy = al.Galaxy(
    bulge=al.EllipticalDevVaucouleurs(
        centre=(0.0, 0.0), axis_ratio=0.9, phi=45.0, intensity=0.1, effective_radius=1.0
    )
)

lens_data = al.LensData(ccd_data=ccd_data, mask=mask)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy])
fit = al.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)
lens_fit_plotters.plot_fit_subplot(
    fit=fit, should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True
)



import numpy as np
import pytest

import autolens as al
from autolens import exc


@pytest.fixture(name="simple_mask_index_array")
def make_simple_mask_index_array():
    return np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])


@pytest.fixture(name="cross_mask")
def make_cross_mask():
    mask = np.full((3, 3), False)

    mask[0, 0] = True
    mask[0, 2] = True
    mask[2, 2] = True
    mask[2, 0] = True

    return mask


@pytest.fixture(name="cross_mask_index_array")
def make_cross_mask_index_array():
    return np.array([[-1, 0, -1], [1, 2, 3], [-1, 4, -1]])


@pytest.fixture(name="simple_image_frame_indexes")
def make_simple_image_frame_indexes(simple_convolver):
    return simple_convolver.make_image_frame_indexes((3, 3))


@pytest.fixture(name="cross_image_frame_indexes")
def make_cross_image_frame_indexes(cross_convolver):
    return cross_convolver.make_image_frame_indexes((3, 3))


@pytest.fixture(name="cross_mask_image_frame_indexes")
def make_cross_mask_image_frame_indexes(cross_convolver):
    return cross_convolver.make_blurring_image_frame_indexes(
        (3, 3), blurring_region_mask=np.full((3, 3), False)
    )


@pytest.fixture(name="simple_convolver")
def make_simple_convolver():
    return al.Convolver(
        mask=np.full((3, 3), False),
        blurring_mask=(np.full((3, 3), False)),
        psf=np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]),
    )


@pytest.fixture(name="cross_convolver")
def make_cross_convolver(cross_mask):
    return al.Convolver(
        mask=cross_mask,
        blurring_mask=(np.full((3, 3), False)),
        psf=np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]),
    )


@pytest.fixture(name="simple_kernel")
def make_simple_kernel():
    return np.array([[0, 0.1, 0], [0.1, 0.6, 0.1], [0, 0.1, 0]])


class TestNumbering(object):
    def test_simple_numbering(self, simple_mask_index_array):
        shape = (3, 3)

        convolver = al.Convolver(
            mask=np.full(shape, False),
            blurring_mask=np.full(shape, False),
            psf=np.ones((1, 1)),
        )

        mask_index_array = convolver.mask_index_array

        assert mask_index_array.shape == shape
        # noinspection PyUnresolvedReferences
        assert (mask_index_array == simple_mask_index_array).all()

    def test_simple_mask(self, cross_mask):
        convolver = al.Convolver(
            mask=cross_mask,
            blurring_mask=np.full(cross_mask.shape, False),
            psf=np.ones((1, 1)),
        )

        assert (
            convolver.mask_index_array
            == np.array([[-1, 0, -1], [1, 2, 3], [-1, 4, -1]])
        ).all()

    def test_even_failure(self):
        with pytest.raises(exc.ConvolutionException):
            al.Convolver(
                mask=np.full((3, 3), False),
                blurring_mask=np.full((3, 3), False),
                psf=np.ones((2, 2)),
            )

    def test_mismatching_masks_failure(self):
        with pytest.raises(exc.ConvolutionException):
            al.Convolver(
                mask=np.full((3, 3), False),
                blurring_mask=np.full((3, 4), False),
                psf=np.ones((1, 1)),
            )


class TestFrameExtraction(object):
    def test_trivial_frame_at_coords(self, simple_convolver):
        frame, psf_frame = simple_convolver.frame_at_coordinates_jit(
            coordinates=(1, 1),
            mask=np.full((3, 3), False),
            mask_index_array=simple_convolver.mask_index_array,
            psf=simple_convolver.psf,
        )

        assert (frame == np.array([i for i in range(9)])).all()

    def test_corner_frame(self, simple_convolver):
        corner_frame = np.array([0, 1, 3, 4, -1, -1, -1, -1, -1])

        frame, psf_frame = simple_convolver.frame_at_coordinates_jit(
            coordinates=(0, 0),
            mask=np.full((3, 3), False),
            mask_index_array=simple_convolver.mask_index_array,
            psf=simple_convolver.psf,
        )

        assert (frame == corner_frame).all()

    def test_trivial_psf_frame_at_coords(self, simple_convolver):
        frame, psf_frame = simple_convolver.frame_at_coordinates_jit(
            coordinates=(1, 1),
            mask=np.full((3, 3), False),
            mask_index_array=simple_convolver.mask_index_array,
            psf=simple_convolver.psf,
        )

        assert (
            psf_frame == np.array([[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]])
        ).all()

    def test_corner_psf_frame(self, simple_convolver):
        corner_frame = [0, 1, 3, 4]

        frame, psf_frame = simple_convolver.frame_at_coordinates_jit(
            coordinates=(0, 0),
            mask=np.full((3, 3), False),
            mask_index_array=simple_convolver.mask_index_array,
            psf=simple_convolver.psf,
        )

        assert (psf_frame == np.array([5.0, 6.0, 8.0, 9.0, -1, -1, -1, -1, -1])).all()

    def test_simple_square(self, simple_convolver):
        assert 9 == len(simple_convolver.image_frame_1d_indexes)

        assert (
            simple_convolver.image_frame_1d_indexes[4]
            == np.array([i for i in range(9)])
        ).all()

    def test_frame_5x5_kernel__at_coords(self):
        convolver = al.Convolver(
            mask=np.full((3, 3), False),
            blurring_mask=(np.full((3, 3), False)),
            psf=np.array(
                [
                    [1.0, 2.0, 3.0, 4.0, 5.0],
                    [6.0, 7.0, 8.0, 9.0, 10.0],
                    [11.0, 12.0, 13.0, 14.0, 15.0],
                    [16.0, 17.0, 18.0, 19.0, 20.0],
                    [21.0, 22.0, 23.0, 24.0, 25.0],
                ]
            ),
        )

        frame, psf_frame = convolver.frame_at_coordinates_jit(
            coordinates=(0, 0),
            mask=np.full((3, 3), False),
            mask_index_array=convolver.mask_index_array,
            psf=convolver.psf,
        )

        assert (
            psf_frame
            == np.array(
                [
                    13.0,
                    14.0,
                    15.0,
                    18.0,
                    19.0,
                    20.0,
                    23.0,
                    24.0,
                    25.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                ]
            )
        ).all()

        frame, psf_frame = convolver.frame_at_coordinates_jit(
            coordinates=(1, 0),
            mask=np.full((3, 3), False),
            mask_index_array=convolver.mask_index_array,
            psf=convolver.psf,
        )

        assert (
            psf_frame
            == np.array(
                [
                    8.0,
                    9.0,
                    10.0,
                    13.0,
                    14.0,
                    15.0,
                    18.0,
                    19.0,
                    20.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1,
                ]
            )
        ).all()

        frame, psf_frame = convolver.frame_at_coordinates_jit(
            coordinates=(1, 1),
            mask=np.full((3, 3), False),
            mask_index_array=convolver.mask_index_array,
            psf=convolver.psf,
        )

        assert (
            psf_frame
            == np.array(
                [
                    7.0,
                    8.0,
                    9.0,
                    12.0,
                    13.0,
                    14.0,
                    17.0,
                    18.0,
                    19.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1,
                ]
            )
        ).all()


class TestImageFrameIndexes(object):
    def test_masked_cross__3x3_kernel(self, cross_convolver):
        assert 5 == len(cross_convolver.image_frame_1d_indexes)

        assert (
            cross_convolver.image_frame_1d_indexes[0]
            == np.array([0, 1, 2, 3, -1, -1, -1, -1, -1])
        ).all()
        assert (
            cross_convolver.image_frame_1d_indexes[1]
            == np.array([0, 1, 2, 4, -1, -1, -1, -1, -1])
        ).all()
        assert (
            cross_convolver.image_frame_1d_indexes[2]
            == np.array([0, 1, 2, 3, 4, -1, -1, -1, -1])
        ).all()
        assert (
            cross_convolver.image_frame_1d_indexes[3]
            == np.array([0, 2, 3, 4, -1, -1, -1, -1, -1])
        ).all()
        assert (
            cross_convolver.image_frame_1d_indexes[4]
            == np.array([1, 2, 3, 4, -1, -1, -1, -1, -1])
        ).all()

    def test_masked_square__3x5_kernel__loses_edge_of_top_and_bottom_rows(self):
        convolver = al.Convolver(
            mask=np.full((3, 3), False),
            blurring_mask=(np.full((3, 3), False)),
            psf=np.ones((3, 5)),
        )

        assert (
            convolver.image_frame_1d_indexes[0]
            == np.array([0, 1, 2, 3, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[1]
            == np.array([0, 1, 2, 3, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[2]
            == np.array([0, 1, 2, 3, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[3]
            == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[4]
            == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[5]
            == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[6]
            == np.array([3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[7]
            == np.array([3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[8]
            == np.array([3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()

    def test_masked_square__5x3_kernel__loses_edge_of_left_and_right_columns(self):
        convolver = al.Convolver(
            mask=np.full((3, 3), False),
            blurring_mask=(np.full((3, 3), False)),
            psf=np.ones((5, 3)),
        )

        assert (
            convolver.image_frame_1d_indexes[0]
            == np.array([0, 1, 3, 4, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[1]
            == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[2]
            == np.array([1, 2, 4, 5, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[3]
            == np.array([0, 1, 3, 4, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[4]
            == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[5]
            == np.array([1, 2, 4, 5, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[6]
            == np.array([0, 1, 3, 4, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[7]
            == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[8]
            == np.array([1, 2, 4, 5, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()

    def test_masked_square__5x5_kernel(self):
        convolver = al.Convolver(
            mask=np.full((3, 3), False),
            blurring_mask=(np.full((3, 3), False)),
            psf=np.ones((5, 5)),
        )

        assert (
            convolver.image_frame_1d_indexes[0]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_indexes[1]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_indexes[2]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_indexes[3]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_indexes[4]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_indexes[5]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_indexes[6]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_indexes[7]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_indexes[8]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()


class TestImageFrameKernels(object):
    def test_simple_square(self, simple_convolver):
        assert 9 == len(simple_convolver.image_frame_1d_indexes)

        assert (
            simple_convolver.image_frame_1d_psfs[4]
            == np.array([[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]])
        ).all()

    def test_masked_square__3x5_kernel__loses_edge_of_top_and_bottom_rows(self):
        convolver = al.Convolver(
            mask=np.full((3, 3), False),
            blurring_mask=(np.full((3, 3), False)),
            psf=np.array(
                [
                    [1.0, 2.0, 3.0, 4.0, 5.0],
                    [6.0, 7.0, 8.0, 9.0, 10.0],
                    [11.0, 12.0, 13.0, 14.0, 15.0],
                ]
            ),
        )

        assert (
            convolver.image_frame_1d_psfs[0]
            == np.array(
                [8.0, 9.0, 10.0, 13.0, 14.0, 15.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_psfs[1]
            == np.array(
                [7.0, 8.0, 9.0, 12.0, 13.0, 14.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_psfs[2]
            == np.array(
                [6.0, 7.0, 8.0, 11.0, 12.0, 13.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_psfs[3]
            == np.array(
                [
                    3.0,
                    4.0,
                    5.0,
                    8.0,
                    9.0,
                    10.0,
                    13.0,
                    14.0,
                    15.0,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_psfs[4]
            == np.array(
                [2.0, 3.0, 4.0, 7.0, 8.0, 9.0, 12.0, 13.0, 14.0, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_psfs[5]
            == np.array(
                [1.0, 2.0, 3.0, 6.0, 7.0, 8.0, 11.0, 12.0, 13.0, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_psfs[6]
            == np.array(
                [3.0, 4.0, 5.0, 8.0, 9.0, 10.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_psfs[7]
            == np.array(
                [2.0, 3.0, 4.0, 7.0, 8.0, 9.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_psfs[8]
            == np.array(
                [1.0, 2.0, 3.0, 6.0, 7.0, 8.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()

    def test_masked_square__5x3_kernel__loses_edge_of_left_and_right_columns(self):
        convolver = al.Convolver(
            mask=np.full((3, 3), False),
            blurring_mask=(np.full((3, 3), False)),
            psf=np.array(
                [
                    [1.0, 2.0, 3.0],
                    [4.0, 5.0, 6.0],
                    [7.0, 8.0, 9.0],
                    [10.0, 11.0, 12.0],
                    [13.0, 14.0, 15.0],
                ]
            ),
        )

        assert (
            convolver.image_frame_1d_psfs[0]
            == np.array(
                [8.0, 9.0, 11.0, 12.0, 14.0, 15.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_psfs[1]
            == np.array(
                [
                    7.0,
                    8.0,
                    9.0,
                    10.0,
                    11.0,
                    12.0,
                    13.0,
                    14.0,
                    15.0,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_psfs[2]
            == np.array(
                [7.0, 8.0, 10.0, 11.0, 13.0, 14.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_psfs[3]
            == np.array(
                [5.0, 6.0, 8.0, 9.0, 11.0, 12.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_psfs[4]
            == np.array(
                [4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_psfs[5]
            == np.array(
                [4.0, 5.0, 7.0, 8.0, 10.0, 11.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_psfs[6]
            == np.array(
                [2.0, 3.0, 5.0, 6.0, 8.0, 9.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_psfs[7]
            == np.array(
                [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_psfs[8]
            == np.array(
                [1.0, 2.0, 4.0, 5.0, 7.0, 8.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()


class TestBlurringFrameIndxes(object):
    def test__blurring_region_3x3_kernel(self, cross_mask):
        blurring_mask = np.array(
            [[False, True, False], [True, True, True], [False, True, False]]
        )

        convolver = al.Convolver(
            mask=cross_mask, blurring_mask=blurring_mask, psf=np.ones((3, 3))
        )

        assert (
            convolver.blurring_frame_1d_indexes[0]
            == np.array([0, 1, 2, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.blurring_frame_1d_indexes[1]
            == np.array([0, 2, 3, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.blurring_frame_1d_indexes[2]
            == np.array([1, 2, 4, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.blurring_frame_1d_indexes[3]
            == np.array([2, 3, 4, -1, -1, -1, -1, -1, -1])
        ).all()

    def test__blurring_region_5x5_kernel(self, cross_mask):
        blurring_mask = np.array(
            [[False, True, False], [True, True, True], [False, True, False]]
        )

        convolver = al.Convolver(
            mask=cross_mask, blurring_mask=blurring_mask, psf=np.ones((5, 5))
        )

        assert (
            convolver.blurring_frame_1d_indexes[0]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.blurring_frame_1d_indexes[1]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.blurring_frame_1d_indexes[2]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.blurring_frame_1d_indexes[3]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()

    def test__convolver_with_blurring_region_added(self, cross_mask):
        blurring_mask = np.array(
            [[False, True, False], [True, True, True], [False, True, False]]
        )

        convolver = al.Convolver(mask=cross_mask, psf=np.ones((3, 3)))

        convolver = convolver.convolver_with_blurring_mask_added(
            blurring_mask=blurring_mask
        )

        assert (
            convolver.blurring_frame_1d_indexes[0]
            == np.array([0, 1, 2, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.blurring_frame_1d_indexes[1]
            == np.array([0, 2, 3, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.blurring_frame_1d_indexes[2]
            == np.array([1, 2, 4, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.blurring_frame_1d_indexes[3]
            == np.array([2, 3, 4, -1, -1, -1, -1, -1, -1])
        ).all()


class TestBlurringFrameKernels(object):
    def test__blurring_region_3x3_kernel(self, cross_mask):
        blurring_mask = np.array(
            [[False, True, False], [True, True, True], [False, True, False]]
        )

        convolver = al.Convolver(
            mask=cross_mask,
            blurring_mask=blurring_mask,
            psf=np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]),
        )

        assert (
            convolver.blurring_frame_1d_psfs[0]
            == np.array([6.0, 8.0, 9.0, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.blurring_frame_1d_psfs[1]
            == np.array([4.0, 7.0, 8.0, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.blurring_frame_1d_psfs[2]
            == np.array([2.0, 3.0, 6.0, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.blurring_frame_1d_psfs[3]
            == np.array([1.0, 2.0, 4.0, -1, -1, -1, -1, -1, -1])
        ).all()

    def test__blurring_region_5x5_kernel(self, cross_mask):
        blurring_mask = np.array(
            [[False, True, False], [True, True, True], [False, True, False]]
        )

        convolver = al.Convolver(
            mask=cross_mask,
            blurring_mask=blurring_mask,
            psf=np.array(
                [
                    [1.0, 2.0, 3.0, 4.0, 5.0],
                    [6.0, 7.0, 8.0, 9.0, 10.0],
                    [11.0, 12.0, 13.0, 14.0, 15.0],
                    [16.0, 17.0, 18.0, 19.0, 20.0],
                    [21.0, 22.0, 23.0, 24.0, 25.0],
                ]
            ),
        )

        assert (
            convolver.blurring_frame_1d_psfs[0]
            == np.array(
                [
                    14.0,
                    18.0,
                    19.0,
                    20.0,
                    24.0,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.blurring_frame_1d_psfs[1]
            == np.array(
                [
                    12.0,
                    16.0,
                    17.0,
                    18.0,
                    22.0,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.blurring_frame_1d_psfs[2]
            == np.array(
                [
                    4.0,
                    8.0,
                    9.0,
                    10.0,
                    14.0,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.blurring_frame_1d_psfs[3]
            == np.array(
                [
                    2.0,
                    6.0,
                    7.0,
                    8.0,
                    12.0,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()


class TestFrameLengths(object):
    def test__frames_are_from_examples_above__lengths_are_right(self):
        convolver = al.Convolver(
            mask=np.full((3, 3), False),
            blurring_mask=(np.full((3, 3), False)),
            psf=np.ones((3, 5)),
        )

        # convolver_image.image_frame_indexes[0] == np.array([0, 1, 2, 3, 4, 5])
        # convolver_image.image_frame_indexes[1] == np.array([0, 1, 2, 3, 4, 5])
        # convolver_image.image_frame_indexes[2] == np.array([0, 1, 2, 3, 4, 5])
        # convolver_image.image_frame_indexes[3] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        # (convolver_image.image_frame_indexes[4] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        # convolver_image.image_frame_indexes[5] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        # convolver_image.image_frame_indexes[6] == np.array([3, 4, 5, 6, 7, 8])
        # convolver_image.image_frame_indexes[7] == np.array([3, 4, 5, 6, 7, 8])
        # convolver_image.image_frame_indexes[8] == np.array([3, 4, 5, 6, 7, 8])

        assert (
            convolver.image_frame_1d_lengths == np.array([6, 6, 6, 9, 9, 9, 6, 6, 6])
        ).all()

    def test__blurring_frames_from_example_above__lengths_are_right(self, cross_mask):
        blurring_mask = np.array(
            [[False, True, False], [True, True, True], [False, True, False]]
        )

        convolver = al.Convolver(
            mask=cross_mask,
            blurring_mask=blurring_mask,
            psf=np.array(
                [
                    [1.0, 2.0, 3.0, 4.0, 5.0],
                    [6.0, 7.0, 8.0, 9.0, 10.0],
                    [11.0, 12.0, 13.0, 14.0, 15.0],
                    [16.0, 17.0, 18.0, 19.0, 20.0],
                    [21.0, 22.0, 23.0, 24.0, 25.0],
                ]
            ),
        )

        # assert (convolver_image.blurring_frame_psfs[0] == np.array([14.0, 18.0, 19.0, 20.0, 24.0])).all()
        # assert (convolver_image.blurring_frame_psfs[1] == np.array([12.0, 16.0, 17.0, 18.0, 22.0])).all()
        # assert (convolver_image.blurring_frame_psfs[2] == np.array([4.0, 8.0, 9.0, 10.0, 14.0])).all()
        # assert (convolver_image.blurring_frame_psfs[3] == np.array([2.0, 6.0, 7.0, 8.0, 12.0])).all()

        assert (convolver.blurring_frame_1d_lengths == np.array([5, 5, 5, 5])).all()


class TestConvolveMappingMatrix(object):
    def test__asymetric_convolver__matrix_blurred_correctly(self):
        shape = (4, 4)
        mask = np.full(shape, False)

        asymmetric_kernel = np.array([[0, 0.0, 0], [0.4, 0.2, 0.3], [0, 0.1, 0]])

        convolver = al.Convolver(mask=mask, psf=asymmetric_kernel)

        mapping = np.array(
            [
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [
                    0,
                    1,
                    0,
                ],  # The 0.3 should be 'chopped' from this pixel as it is on the right-most edge
                [0, 0, 0],
                [1, 0, 0],
                [0, 0, 1],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
            ]
        )

        blurred_mapping = convolver.convolve_mapping_matrix(mapping)

        assert (
            blurred_mapping
            == np.array(
                [
                    [0, 0, 0],
                    [0, 0, 0],
                    [0, 0, 0],
                    [0, 0, 0],
                    [0, 0, 0],
                    [0, 0, 0],
                    [0, 0.4, 0],
                    [0, 0.2, 0],
                    [0.4, 0, 0],
                    [0.2, 0, 0.4],
                    [0.3, 0, 0.2],
                    [0, 0.1, 0.3],
                    [0, 0, 0],
                    [0.1, 0, 0],
                    [0, 0, 0.1],
                    [0, 0, 0],
                ]
            )
        ).all()

    def test__asymetric_convolver__multiple_overlapping_blurred_entires_in_matrix(self):
        shape = (4, 4)
        mask = np.full(shape, False)

        asymmetric_kernel = np.array([[0, 0.0, 0], [0.4, 0.2, 0.3], [0, 0.1, 0]])

        convolver = al.Convolver(mask=mask, psf=asymmetric_kernel)

        mapping = np.array(
            [
                [0, 1, 0],
                [0, 1, 0],
                [0, 1, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [
                    0,
                    1,
                    0,
                ],  # The 0.3 should be 'chopped' from this pixel as it is on the right-most edge
                [1, 0, 0],
                [1, 0, 0],
                [0, 0, 1],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
            ]
        )

        blurred_mapping = convolver.convolve_mapping_matrix(mapping)

        assert blurred_mapping == pytest.approx(
            np.array(
                [
                    [0, 0.6, 0],
                    [0, 0.9, 0],
                    [0, 0.5, 0],
                    [0, 0.3, 0],
                    [0, 0.1, 0],
                    [0, 0.1, 0],
                    [0, 0.5, 0],
                    [0, 0.2, 0],
                    [0.6, 0, 0],
                    [0.5, 0, 0.4],
                    [0.3, 0, 0.2],
                    [0, 0.1, 0.3],
                    [0.1, 0, 0],
                    [0.1, 0, 0],
                    [0, 0, 0.1],
                    [0, 0, 0],
                ]
            ),
            1e-4,
        )


class TestConvolution(object):
    def test_cross_mask_with_blurring_entries(self, cross_mask):
        kernel = np.array([[0, 0.2, 0], [0.2, 0.4, 0.2], [0, 0.2, 0]])

        blurring_mask = np.array(
            [[False, True, False], [True, True, True], [False, True, False]]
        )

        convolver = al.Convolver(
            mask=cross_mask, blurring_mask=blurring_mask, psf=kernel
        )

        pixel_array = np.array([1, 0, 0, 0, 0])
        blurring_array = np.array([1, 0, 0, 0])

        result = convolver.convolve_image(pixel_array, blurring_array)

        assert (np.round(result, 1) == np.array([0.6, 0.2, 0.2, 0.0, 0.0])).all()


class TestCompareToFull2dConv:
    def test__compare_convolver_to_2d_convolution(self):
        # Setup a blurred datas_, using the PSF to perform the convolution in 2D, then masks it to make a 1d array.

        im = np.arange(900).reshape(30, 30)
        psf = al.PSF(array=np.arange(49).reshape(7, 7), pixel_scale=1.0)
        blurred_im = psf.convolve(im)
        msk = al.Mask.circular(shape=(30, 30), pixel_scale=1.0, radius_arcsec=4.0)
        blurred_masked_im_0 = msk.array_1d_from_array_2d(blurred_im)

        # Now reproduce this datas_ using the frame convolver_image

        blurring_mask = msk.blurring_mask_from_psf_shape(psf.shape)
        convolver = al.Convolver(mask=msk, blurring_mask=blurring_mask, psf=psf)
        im_1d = msk.array_1d_from_array_2d(im)
        blurring_im_1d = blurring_mask.array_1d_from_array_2d(im)
        blurred_masked_im_1 = convolver.convolve_image(
            image_array=im_1d, blurring_array=blurring_im_1d
        )

        assert blurred_masked_im_0 == pytest.approx(blurred_masked_im_1, 1e-4)
